#!/usr/bin/env python3


"""Calculate detailed itineraries between many origins and destinations."""


import copy
import warnings

try:
    from warnings import deprecated
except ImportError:  # Python<=3.12
    from typing_extensions import deprecated

import geopandas
import joblib
import pandas

from .base_travel_time_matrix import BaseTravelTimeMatrix
from .trip import Trip
from .trip_planner import ACCURATE_GEOMETRIES, TripPlanner


__all__ = ["DetailedItineraries", "DetailedItinerariesComputer"]


class DetailedItineraries(BaseTravelTimeMatrix):
    """Compute detailed itineraries between many origins and destinations."""

    COLUMNS = ["from_id", "to_id", "option"] + Trip.COLUMNS

    _r5py_attributes = BaseTravelTimeMatrix._r5py_attributes + [
        "all_to_all",
        "od_pairs",
    ]

    def __init__(
        self,
        transport_network,
        origins=None,
        destinations=None,
        snap_to_network=False,
        force_all_to_all=False,
        **kwargs,
    ):
        """
        Compute travel times between many origins and destinations.

        ``r5py.DetailedItineraries`` are child classes of
        ``geopandas.GeoDataFrame`` and support all of their methods and
        properties, see https://geopandas.org/en/stable/docs.html

        Arguments
        ---------
        transport_network : r5py.TransportNetwork | tuple(str, list(str), dict)
            The transport network to route on. This can either be a readily
            initialised r5py.TransportNetwork or a tuple of the parameters
            passed to ``TransportNetwork.__init__()``: the path to an OpenStreetMap
            extract in PBF format, a list of zero of more paths to GTFS transport
            schedule files, and a dict with ``build_config`` options.
        origins : geopandas.GeoDataFrame
            Places to find a route _from_
            Has to have a point geometry, and at least an `id` column
        destinations : geopandas.GeoDataFrame (optional)
            Places to find a route _to_
            Has to have a point geometry, and at least an `id` column
            If omitted, use same data set as for origins
        snap_to_network : bool or int, default False
            Should origin an destination points be snapped to the street network
            before routing? If `True`, the default search radius (defined in
            `com.conveyal.r5.streets.StreetLayer.LINK_RADIUS_METERS`) is used,
            if `int`, use `snap_to_network` meters as the search radius.
        force_all_to_all : bool, default False
            If ``origins`` and ``destinations`` have the same length, by
            default, ``DetailedItineraries`` finds routes between pairs
            of origins and destinations, i.e., it routes from origin #1 to
            destination #1, origin #2 to destination #2, ... .
            Set ``force_all_to_all=True`` to route from each origin to all
            destinations (this is the default, if ``origins`` and ``destinations``
            have different lengths, or if ``destinations`` is omitted)
        **kwargs : mixed
            Any arguments than can be passed to r5py.RegionalTask:
            ``departure``, ``departure_time_window``, ``percentiles``, ``transport_modes``,
            ``access_modes``, ``egress_modes``, ``max_time``, ``max_time_walking``,
            ``max_time_cycling``, ``max_time_driving``, ``speed_cycling``, ``speed_walking``,
            ``max_public_transport_rides``, ``max_bicycle_traffic_stress``
            Note that not all arguments might make sense in this context, and the
            underlying R5 engine might ignore some of them.
        """
        super().__init__(
            transport_network,
            origins,
            destinations,
            snap_to_network,
            **kwargs,
        )

        if destinations is None:
            self.all_to_all = True
            if self.verbose:
                warnings.warn(
                    "No destinations specified, computing an all-to-all matrix",
                    RuntimeWarning,
                )

        elif len(origins) != len(destinations):
            self.all_to_all = True
            if self.verbose:
                warnings.warn(
                    "Origins and destinations are of different length, computing an all-to-all matrix",
                    RuntimeWarning,
                )
        elif origins.equals(destinations):
            self.all_to_all = True
            if self.verbose:
                warnings.warn(
                    "Origins and destinations are identical, computing an all-to-all matrix",
                    RuntimeWarning,
                )
        else:
            self.all_to_all = force_all_to_all

        data = self._compute()
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=(
                    "You are adding a column named 'geometry' to a GeoDataFrame "
                    "constructed without an active geometry column"
                ),
                category=FutureWarning,
            )
            for column in data.columns:
                self[column] = data[column]
            self.set_geometry("geometry")

        del self.transport_network

    def _compute(self):
        """
        Compute travel times from all origins to all destinations.

        Returns
        -------
        geopandas.GeoDataFrame
            The resulting detailed routes. For each origin/destination pair,
            multiple route alternatives (‘options’) might be reported that each
            consist of one or more segments. Each segment represents one row.
            multiple route alternatives (‘options’) might be reported that each consist of
            one or more segments. Each segment represents one row.

            The data frame comprises of the following columns: `from_id`,
            `to_id`, `option` (`int`), `segment` (`int`), `transport_mode`
            (`r5py.TransportMode`), `departure_time` (`datetime.datetime`),
            `distance` (`float`, metres), `travel_time` (`datetime.timedelta`),
            `wait_time` (`datetime.timedelta`), `feed` (`str`, the feed name
            used), `agency_id` (`str` the public transport agency identifier),
            `route_id` (`str`, public transport route ID), `start_stop_id`
            (`str`, the GTFS stop_id for boarding), `end_stop_id` (`str`, the
            GTFS stop_id for alighting), `geometry` (`shapely.LineString`)
        """
        self._prepare_origins_destinations()

        # warn if public transport routes are requested, but R5 has been
        # compiled with `TransitLayer.SAVE_SHAPES = false`.
        if [
            mode for mode in self.request.transport_modes if mode.is_transit_mode
        ] and not ACCURATE_GEOMETRIES:
            warnings.warn(
                (
                    "R5 has been compiled with "
                    "`TransitLayer.SAVE_SHAPES = false` (the default). "
                    "The geometries of public transport routes are "
                    "inaccurate (straight lines between stops), and "
                    "distances can not be computed."
                ),
                RuntimeWarning,
            )

        # loop over all origin/destination pairs, modify the request, and
        # compute times, distance, and other details for each trip
        with joblib.Parallel(
            prefer="threads",
            verbose=(10 * self.verbose),  # joblib has a funny verbosity scale
            n_jobs=self.NUM_THREADS,
        ) as parallel:
            matrices = parallel(
                joblib.delayed(self._travel_details_per_od_pair)(from_id, to_id)
                for _, (from_id, to_id) in self.od_pairs.iterrows()
            )
            od_matrix = pandas.concat(
                [matrix.astype(matrices[0].dtypes) for matrix in matrices],
                ignore_index=True,
            )

        od_matrix = geopandas.GeoDataFrame(od_matrix, crs=self._origins_crs)
        return od_matrix

    def _prepare_origins_destinations(self):
        """Make sure we received enough information to route from origins to destinations."""
        super()._prepare_origins_destinations()

        if self.all_to_all:
            # manually create a list of all all-to-all permutations
            self.od_pairs = self.origins[["id"]].join(
                self.destinations[["id"]],
                how="cross",
                lsuffix="_origin",
                rsuffix="_destination",
            )
        else:
            # origins and destinations are same length, run one-to-one routing
            self.od_pairs = pandas.DataFrame(
                {
                    "id_origin": self.origins.id.values,
                    "id_destination": self.destinations.id.values,
                }
            )

    def _travel_details_per_od_pair(self, from_id, to_id):
        origin = self.origins[self.origins.id == from_id]
        destination = self.destinations[self.destinations.id == to_id]

        request = copy.copy(self.request)
        request._regional_task.fromLat = origin.geometry.item().y
        request._regional_task.fromLon = origin.geometry.item().x
        request._regional_task.toLat = destination.geometry.item().y
        request._regional_task.toLon = destination.geometry.item().x

        trip_planner = TripPlanner(self.transport_network, request)
        trips = trip_planner.trips

        # fmt: off
        trips = [
            [from_id, to_id, option] + segment
            for option, trip in enumerate(trips)
            for segment in trip.as_table()
        ]
        # fmt: on

        return pandas.DataFrame(trips, columns=self.COLUMNS)


@deprecated(
    "Use `DetailedItineraries` instead, `DetailedItinerariesComputer will be deprecated in a future release."
)
class DetailedItinerariesComputer:
    """Compute detailed itineraries between many origins and destinations."""

    def __init__(self, *args, **kwargs):
        """Compute detailed itineraries between many origins and destinations."""
        self._detailed_itineraries = DetailedItineraries(*args, **kwargs)

    def compute_travel_details(self):
        """
        Compute travel times from all origins to all destinations.

        Returns
        -------
        geopandas.GeoDataFrame
            The resulting detailed routes. For each origin/destination pair,
            multiple route alternatives (‘options’) might be reported that each consist of
            one or more segments. Each segment represents one row.
            The data frame comprises of the following columns: `from_id`,
            `to_id`, `option` (`int`), `segment` (`int`), `transport_mode`
            (`r5py.TransportMode`), `departure_time` (`datetime.datetime`),
            `distance` (`float`, metres), `travel_time` (`datetime.timedelta`),
            `wait_time` (`datetime.timedelta`), `route` (`str`, public transport
            route number or name), `geometry` (`shapely.LineString`)
            TODO: Add description of output data frame columns and format
        """
        return self._detailed_itineraries



#!/usr/bin/env python3

"""Wraps a com.conveyal.r5.analyst.cluster.RegionalTask."""

import collections.abc
import datetime
import warnings

import jpype

from .scenario import Scenario
from .transport_mode import TransportMode
from ..util import start_jvm

import java.io
import java.time
import com.conveyal.r5


__all__ = ["RegionalTask"]


start_jvm()


class RegionalTask:
    """Create a RegionalTask, a computing request for R5."""

    def __init__(
        self,
        transport_network,
        origin=None,
        destinations=None,
        departure=datetime.datetime.now(),
        departure_time_window=datetime.timedelta(minutes=10),
        percentiles=[50],
        transport_modes=[TransportMode.TRANSIT],
        access_modes=[TransportMode.WALK],
        egress_modes=None,  # default: access_modes
        max_time=datetime.timedelta(hours=2),
        max_time_walking=datetime.timedelta(hours=2),
        max_time_cycling=datetime.timedelta(hours=2),
        max_time_driving=datetime.timedelta(hours=2),
        speed_walking=3.6,
        speed_cycling=12.0,
        max_public_transport_rides=8,
        max_bicycle_traffic_stress=3,
    ):
        """
        Create a RegionalTask, a computing request for R5.

        A RegionalTask wraps a `com.conveyal.r5.analyst.cluster.RegionalTask`,
        which is used to specify the details of a requested computation.
        RegionalTasks underlie virtually all major computations carried out,
        such as, e.g., `TravelTimeMatrix` or `AccessibilityEstimator`.

        In **r5py**, there is usually no need to explicitely create a
        `RegionalTask`. Rather, the constructors to the computation classes
        (`TravelTimeMatrix`, `AccessibilityEstimator`, ...) accept the
        arguments, and pass them through to an internally handled
        `RegionalTask`.

        Arguments
        ---------
        transport_network : r5py.TransportNetwork
            The street + public transport network to route on
        origin : shapely.geometry.Point
            Point to route from
        destinations : geopandas.GeoDataFrame
            Points to route to, has to have at least an ``id`` column and a
            geometry
        departure : datetime.datetime
            Find public transport connections leaving every minute within
            ``departure_time_window`` after ``departure``. Default: current date
            and time
        departure_time_window : datetime.timedelta
            (see ``departure``) Default: 10 minutes
        percentiles : list[int]
            Return the travel time for these percentiles of all computed trips,
            by travel time. By default, return the median travel time. Default:
            [50]
        transport_modes : list[r5py.TransportMode] or list[str]
            The mode of transport to use for routing. Can be a r5py mode
            enumerable, or a string representation (e.g. "TRANSIT") Default:
            [r5py.TransportMode.TRANSIT] (all public transport)
        access_modes : list[r5py.TransportMode] or list[str]
            Mode of transport to public transport stops. Can be a r5py mode
            object, or a string representation (e.g. "WALK") Default:
            [r5py.TransportMode.WALK]
        egress_modes : list[r5py.TransportMode]
            Mode of transport from public transport stops. Default: access_modes
        max_time : datetime.timedelta
            Maximum trip duration. Default: 2 hours
        max_time_walking : datetime.timedelta
            Maximum time spent walking, potentially including access and egress
            Default: max_time
        max_time_cycling : datetime.timedelta
            Maximum time spent cycling, potentially including access and egress
            Default: max_time
        max_time_driving : datetime.timedelta
            Maximum time spent driving Default: max_time
        speed_walking : float
            Mean walking speed for routing, km/h. Default: 3.6 km/h
        speed_cycling : float
            Mean cycling speed for routing, km/h. Default: 12.0 km/h
        max_public_transport_rides : int
            Use at most ``max_public_transport_rides`` consecutive public
            transport connections. Default: 8
        max_bicycle_traffic_stress : int
            Maximum stress level for cyclist routing, ranges from 1-4 see
            https://docs.conveyal.com/learn-more/traffic-stress Default: 3
        """
        self._regional_task = com.conveyal.r5.analyst.cluster.RegionalTask()
        self.scenario = Scenario()

        self.transport_network = transport_network

        self.origin = origin
        self.destinations = destinations

        self.access_modes = access_modes
        self.egress_modes = egress_modes if egress_modes is not None else access_modes
        # last, because extra logic that depends on the others having been set
        self.transport_modes = transport_modes

        self.departure = departure  # transport modes should be set before this line
        self.departure_time_window = departure_time_window
        self.percentiles = percentiles

        self.max_time = max_time
        self.max_time_walking = (
            max_time_walking if max_time_walking is not None else max_time
        )
        self.max_time_cycling = (
            max_time_cycling if max_time_cycling is not None else max_time
        )
        self.max_time_driving = (
            max_time_driving if max_time_driving is not None else max_time
        )

        self.speed_cycling = speed_cycling
        self.speed_walking = speed_walking

        self.max_public_transport_rides = max_public_transport_rides
        self.max_bicycle_traffic_stress = max_bicycle_traffic_stress

        # always record travel times
        self._regional_task.recordTimes = True

        # a few settings we don’t expose (yet?)
        self._regional_task.makeTauiSite = False
        self._regional_task.oneToOne = False
        self._regional_task.monteCarloDraws = 60
        self._regional_task.recordAccessibility = False

    def __copy__(self):
        """Override `copy.copy()` to properly handle Java classes."""
        # remember all __dict__ except `_regional_task`, which is a Java `Clonable`
        _dict = {k: v for k, v in self.__dict__.items() if k != "_regional_task"}
        # add a clone of the Java object
        _dict["_regional_task"] = self._regional_task.clone()

        # create a new instance, copy over the __dict__ created above
        clone = super().__new__(self.__class__)
        clone.__dict__ = _dict

        return clone

    @property
    def access_modes(self):
        """Route with these modes of transport to reach public transport (r5py.TransportMode)."""
        return self._access_modes

    @access_modes.setter
    def access_modes(self, access_modes):
        # eliminate duplicates, cast to TransportMode (converts str values)
        access_modes = set([TransportMode(mode) for mode in set(access_modes)])
        self._access_modes = access_modes
        self._regional_task.accessModes = RegionalTask._enum_set(
            access_modes, com.conveyal.r5.api.util.LegMode
        )

    @property
    def departure(self):
        """Find public transport connections leaving within ``departure_time_window`` after ``departure`` (datetime.datetime)."""
        return self._departure

    @departure.setter
    def departure(self, departure):
        # fmt: off
        if (
            [mode for mode in self.transport_modes if mode.is_transit_mode]
            and not self.transport_network.transit_layer.covers(departure)
        ):
            # fmt: on
            warnings.warn(
                f"Departure time {departure} is outside of the time range "
                "covered by currently loaded GTFS data sets.",
                RuntimeWarning,
            )

        self._departure = departure
        self._regional_task.date = java.time.LocalDate.of(
            departure.year, departure.month, departure.day
        )
        # seconds from midnight
        self._regional_task.fromTime = int(
            datetime.timedelta(
                hours=departure.hour, minutes=departure.minute
            ).total_seconds()
        )
        try:
            self._regional_task.toTime = int(
                self._regional_task.fromTime
                + self.departure_time_window.total_seconds()
            )
        except AttributeError:  # departure_time_window has not been set yet
            pass

    @property
    def departure_time_window(self):
        """Find public transport connections leaving within ``departure_time_window`` after ``departure`` (datetime.timedelta).

        **Note:** The value of ``departure_time_window`` should be set with some
        caution. Specifically, setting values near or below the typical headways
        in the studied transit network may lead to routing problems. See `this
        GitHub discussion <https://github.com/r5py/r5py/issues/292>`_ for
        details.
        """
        return self._departure_time_window

    @departure_time_window.setter
    def departure_time_window(self, departure_time_window: datetime.timedelta):
        if departure_time_window.total_seconds() < 300:
            warnings.warn(
                "The provided departure time window is below 5 minutes. This may cause adverse effects with routing.",
                RuntimeWarning,
            )
        self._departure_time_window = departure_time_window
        self._regional_task.toTime = int(
            self._regional_task.fromTime + departure_time_window.total_seconds()
        )

    @property
    def destinations(self):
        """
        Points to route to.

        A ``geopandas.GeoDataFrame`` with a point geometry, and at least
        an ``id`` column (which R5 mangles to ``str``).
        """
        return self._destinations

    @destinations.setter
    def destinations(self, destinations):
        if destinations is not None:
            self._destinations = destinations

            # wrap destinations in a few layers of streams (yeah, Java)
            output_stream = java.io.ByteArrayOutputStream()
            data_output_stream = java.io.DataOutputStream(output_stream)

            # first: number of destinations
            data_output_stream.writeInt(len(destinations))

            # then, data columns, one by one, then still ‘opportunties’
            for id_ in destinations.id.astype(str):
                data_output_stream.writeUTF(id_)
            for lat in destinations.geometry.y:
                data_output_stream.writeDouble(lat)
            for lon in destinations.geometry.x:
                data_output_stream.writeDouble(lon)
            for _ in range(len(destinations)):
                data_output_stream.writeDouble(0)  # ‘opportunities’

            # convert to input stream, then into a point set
            destinations_point_set = com.conveyal.r5.analyst.FreeFormPointSet(
                java.io.ByteArrayInputStream(output_stream.toByteArray())
            )

            self._regional_task.destinationPointSets = [destinations_point_set]

            # TODO: figure out whether we could cut this a bit shorter. We should be able
            # to construct the ByteArray fed to java.io.ByteArrayInputStream as a Python `bytes`
            # without the detour via two Java OutputStreams.
            # (but not sure how to distinguish between the writeUTF/writeDouble/etc)

    @property
    def egress_modes(self):
        """Route with these modes of transport to reach the destination from public transport (r5py.TransportMode)."""
        return self._egress_modes

    @egress_modes.setter
    def egress_modes(self, egress_modes):
        # eliminate duplicates, cast to TransportMode (converts str values)
        egress_modes = set([TransportMode(mode) for mode in set(egress_modes)])
        self._egress_modes = egress_modes
        self._regional_task.egressModes = RegionalTask._enum_set(
            egress_modes, com.conveyal.r5.api.util.LegMode
        )

    @property
    def max_bicycle_traffic_stress(self):
        """
        Find routes with this maximum stress level for cyclists.

        Int, in the range 1-4, see https://docs.conveyal.com/learn-more/traffic-stress
        """
        return self._max_bicycle_traffic_stress

    @max_bicycle_traffic_stress.setter
    def max_bicycle_traffic_stress(self, max_bicycle_traffic_stress):
        self._max_bicycle_traffic_stress = max_bicycle_traffic_stress
        self._regional_task.bikeTrafficStress = max_bicycle_traffic_stress

    @property
    def max_public_transport_rides(self):
        """Include at most this many consecutive public transport rides (int)."""
        return self._max_public_transport_rides

    @max_public_transport_rides.setter
    def max_public_transport_rides(self, max_public_transport_rides):
        self._max_public_transport_rides = max_public_transport_rides
        self._regional_task.maxRides = max_public_transport_rides

    @property
    def max_time(self):
        """Restrict trip duration (datetime.timedelta)."""
        return self._max_time

    @max_time.setter
    def max_time(self, max_time):
        self._max_time = max_time
        max_time = int(max_time.total_seconds() / 60)
        self._regional_task.streetTime = max_time
        self._regional_task.maxTripDurationMinutes = max_time
        self._regional_task.maxCarTime = max_time

    @property
    def max_time_cycling(self):
        """
        Restrict routes to at most this duration of cycling (datetime.timedelta).

        Depending on the transport modes specified, this includes times
        on the main leg of the trip, as well as during access and egress.
        """
        return self._max_time_cycling

    @max_time_cycling.setter
    def max_time_cycling(self, max_time_cycling):
        self._max_time_cycling = max_time_cycling
        self._regional_task.maxBikeTime = int(max_time_cycling.total_seconds() / 60)

    @property
    def max_time_driving(self):
        """Restrict routes to at most this duration of driving (datetime.timedelta)."""
        return self._max_time_driving

    @max_time_driving.setter
    def max_time_driving(self, max_time_driving):
        self._max_time_driving = max_time_driving
        self._regional_task.maxCarTime = int(max_time_driving.total_seconds() / 60)

    @property
    def max_time_walking(self):
        """
        Restrict routes to at most this duration of walking (datetime.timedelta).

        Depending on the transport modes specified, this includes times
        on the main leg of the trip, as well as during access and egress.
        """
        return self._max_time_walking

    @max_time_walking.setter
    def max_time_walking(self, max_time_walking):
        self._max_time_walking = max_time_walking
        self._regional_task.maxWalkTime = int(max_time_walking.total_seconds() / 60)

    @property
    def percentiles(self):
        """
        Return the travel time for these percentiles of all computed trips, by travel time.

        By default, return the median travel time.
        (collections.abc.Sequence[int])
        """
        return self._percentiles

    @percentiles.setter
    def percentiles(self, percentiles):
        try:
            assert isinstance(percentiles, collections.abc.Sequence)
            assert len(percentiles) <= 5  # R5 does not allow more than five percentiles
            # (compare https://github.com/r5py/r5py/issues/139 )
        except AssertionError as exception:
            raise ValueError(
                "Maximum number of percentiles allowed is 5"
            ) from exception
        self._percentiles = percentiles
        self._regional_task.percentiles = percentiles

    # TODO: implement a proper balancing mechanism between the different per-mode
    # maximum times, i.e., a sanity check that the different more specific max_times
    # don’t exceed max_time, for instance, but probably also more complex interrelations
    # (this needs some sitting down with pen and paper and a large cup of tea)

    @property
    def origin(self):
        """Set the origin for the routing operation (shapely.geometry.Point)."""
        return self._origin

    @origin.setter
    def origin(self, origin):
        """
        Set origin geometry.

        Arguments:
        ----------
        origin : shapely.geometry.Point
            Point to route from
        """
        if origin is not None:
            self._origin = origin
            self._regional_task.fromLat = origin.y
            self._regional_task.fromLon = origin.x

    @property
    def scenario(self):
        """Expose the ``RegionalTask``’s ``Scenario`` to Python."""
        return self._regional_task.scenario

    @scenario.setter
    def scenario(self, scenario):
        self._regional_task.scenario = scenario
        self._regional_task.scenarioId = scenario.id

    @property
    def speed_cycling(self):
        """Use this speed for routing for cyclists (km/h, float)."""
        return self._speed_cycling

    @speed_cycling.setter
    def speed_cycling(self, speed_cycling):
        self._speed_cycling = speed_cycling
        self._regional_task.bikeSpeed = speed_cycling / 3600 * 1000  # km/h -> m/s

    @property
    def speed_walking(self):
        """Use this speed for routing pedestrian movement (km/h, float)."""
        return self._speed_walking

    @speed_walking.setter
    def speed_walking(self, speed_walking):
        self._speed_walking = speed_walking
        self._regional_task.walkSpeed = speed_walking / 3600 * 1000  # km/h -> m/s

    @property
    def transport_modes(self):
        """
        Get/set the transport modes used to route the main leg of trips.

        (list[r5py.TransportMode])
        """
        return self._transport_modes

    @transport_modes.setter
    def transport_modes(self, transport_modes):
        # eliminate duplicates, cast to TransportMode (converts str values)
        transport_modes = set([TransportMode(mode) for mode in set(transport_modes)])
        self._transport_modes = transport_modes

        # split them up into direct and transit modes,
        transit_modes = [mode for mode in transport_modes if mode.is_transit_mode]
        direct_modes = [mode for mode in transport_modes if mode.is_street_mode]

        # the different modes underlie certain rules
        # e.g., some direct modes require certain access modes
        # see https://github.com/ipeaGIT/r5r/blob/2e8b9acfd81834f185d95ce53dc5c34beb1315f2/r-package/R/utils.R#L86
        if transit_modes:  # public transport:
            egress_modes = self.egress_modes
            if TransportMode.TRANSIT in transport_modes:
                transit_modes = [mode for mode in TransportMode if mode.is_transit_mode]
            if not direct_modes:
                # only public transport modes passed in,
                # let people walk to and from the stops
                access_modes = direct_modes = [TransportMode.WALK]
            else:
                # otherwise, include the supplied direct modes into access_modes
                access_modes = set(list(self.access_modes) + direct_modes)

        else:  # no public transport
            egress_modes = []  # ignore egress (why?)

            #     # this is weird (the following is the logic implemented in r5r)
            #     # I reckon this is trying to keep the fastest mode only, and
            #     # assumes that car is always faster that bike is always faster than walking
            #     if TransportMode.CAR in transport_modes:
            #         access_modes = direct_modes = [TransportMode.CAR]
            #     elif TransportMode.BICYCLE in transport_modes:
            #         access_modes = direct_modes = [TransportMode.BICYCLE]
            #     elif TransportMode.WALK in transport_modes:
            #         access_modes = direct_modes = [TransportMode.WALK]

            # let’s do that differently (even if potentially more expensive, computationally)
            access_modes = direct_modes

        self.access_modes = access_modes
        self.egress_modes = egress_modes

        self._regional_task.transitModes = RegionalTask._enum_set(
            transit_modes, com.conveyal.r5.api.util.TransitModes
        )
        self._regional_task.directModes = RegionalTask._enum_set(
            direct_modes, com.conveyal.r5.api.util.LegMode
        )

        # pre-compute closest road segments/public transport stops to destination points
        # (for fully-interconnected travel time matrices this also covers all origin points,
        # but potentially this needs to be extended to run also for origins) //TODO

        if self._regional_task.destinationPointSets is not None:
            for mode in direct_modes:
                for destination_point_set in self._regional_task.destinationPointSets:
                    self.transport_network.linkage_cache.getLinkage(
                        destination_point_set,
                        self.transport_network.street_layer,
                        mode,
                    )

    @staticmethod
    def _enum_set(values, java_class):
        # helper function to construct a Java EnumSet out of a list of enum.Enum
        enum_set = java.util.EnumSet.noneOf(java_class)
        for mode in values:
            enum_set.add(java_class.valueOf(mode.value))
        return enum_set


@jpype._jcustomizer.JConversion(
    "com.conveyal.r5.analyst.cluster.AnalysisWorkerTask", exact=RegionalTask
)
@jpype._jcustomizer.JConversion(
    "com.conveyal.r5.profile.ProfileRequest", exact=RegionalTask
)
@jpype._jcustomizer.JConversion(
    "com.conveyal.r5.analyst.cluster.RegionalTask", exact=RegionalTask
)
def _cast_RegionalTask(java_class, object_):
    return object_._regional_task.clone()
    # cloned, so we can reuse the Python instance (e.g., with next origin)



#!/usr/bin/env python3

"""R5py-specific exceptions."""


import requests


# base class for r5py-related exceptions
class R5pyError(Exception):
    """Generic base exception for r5py errors."""


# more specific exceptions
class ChecksumFailed(requests.RequestException, R5pyError):
    """Requested resource did not pass checksum test."""


class MissingColumnError(ValueError, R5pyError):
    """An input data set is missing a required column."""


class NoCrsError(ValueError, R5pyError):
    """An input data set’s geometry column does not have a reference system defined."""


class NonUniqueIDError(ValueError, R5pyError):
    """An input data set’s `id` column has non-unique values."""


class NoIDColumnError(MissingColumnError):
    """An input data set does not have a required `id` column."""


class UnexpectedClasspathSchema(ValueError, R5pyError):
    """A classpath was supplied as an URI, but could not be parsed."""


class UnexpectedCrsError(ValueError, R5pyError):
    """A geometry is in an unexpected reference system."""



#!/usr/bin/env python3


import pytest

import r5py.util


class TestCamelToSnakeCase:
    @pytest.mark.parametrize(
        ["camel_case", "snake_case"],
        [
            ("camelCase", "camel_case"),
            ("CamelCase", "camel_case"),
            ("getHTTPResponseCode", "get_http_response_code"),
            ("HTTPResponseCodeXYZ", "http_response_code_xyz"),
        ],
    )
    def test_case_camel_case(self, camel_case, snake_case):
        assert r5py.util.camel_to_snake_case(camel_case) == snake_case



#!/usr/bin/env python3


import r5py


class TestTripPlanner:
    def test_trip_planner_initialization(
        self,
        transport_network,
        regional_task,
    ):
        _ = r5py.r5.trip_planner.TripPlanner(transport_network, regional_task)



#!/usr/bin/env python3


"""An auto-use fixture that calls Java garbage collection before every function."""


import jpype
import pytest


@pytest.fixture(autouse=True, scope="function")
def java_garbage_collection():
    """Call Java GC before every function."""
    jpype.java.lang.System.gc()



#!/usr/bin/env python3


import geopandas
import pandas
import pytest
import pytest_lazy_fixtures
import shapely

import r5py


class TestDataValidation:
    @pytest.fixture
    def geodataframe_with_all_required_columns(self):
        df = geopandas.GeoDataFrame(
            {
                "id": [1],
                "geometry": [shapely.Point()],
            },
            crs="EPSG:4326",
        )
        yield df

    @pytest.fixture
    def geodataframe_without_id_column(self, geodataframe_with_all_required_columns):
        yield geodataframe_with_all_required_columns[["geometry"]]

    @pytest.fixture
    def geodataframe_with_nonunique_id_column(
        self, geodataframe_with_all_required_columns
    ):
        yield pandas.concat(
            [
                geodataframe_with_all_required_columns,
                geodataframe_with_all_required_columns,
            ]
        )

    @pytest.fixture
    def geodataframe_without_crs(self, geodataframe_with_all_required_columns):
        df = geopandas.GeoDataFrame(
            {
                "id": [1],
                "geometry": [shapely.Point()],
            }
        )
        yield df

    @pytest.mark.parametrize(
        ["geodataframe", "expected_exception"],
        [
            (pytest_lazy_fixtures.lf("geodataframe_with_all_required_columns"), False),
            (
                pytest_lazy_fixtures.lf("geodataframe_without_id_column"),
                r5py.util.exceptions.NoIDColumnError,
            ),
            (
                pytest_lazy_fixtures.lf("geodataframe_with_nonunique_id_column"),
                r5py.util.exceptions.NonUniqueIDError,
            ),
            (
                pytest_lazy_fixtures.lf("geodataframe_without_crs"),
                r5py.util.exceptions.NoCrsError,
            ),
        ],
    )
    def test_check_od_data_set(self, geodataframe, expected_exception):
        if expected_exception:
            with pytest.raises(expected_exception):
                r5py.util.data_validation.check_od_data_set(geodataframe)
        else:
            r5py.util.data_validation.check_od_data_set(geodataframe)



#!/usr/bin/env python3

import datetime

import pytest
import shapely

import r5py


class TestTrip:
    def test_str(self):
        trip_leg = r5py.r5.trip_leg.TripLeg(
            transport_mode=r5py.TransportMode.TRANSIT,
            departure_time=datetime.datetime(2023, 4, 25, 15, 30),
            distance=12.67,
            travel_time=datetime.timedelta(minutes=10),
            wait_time=datetime.timedelta(minutes=2),
            route_id="48A",
            geometry=shapely.LineString([[0, 0], [1, 1], [2, 2]]),
        )
        trip = r5py.r5.trip.Trip([trip_leg])
        assert (
            repr(trip)
            == "<Trip: 12.67m, 600.0s, <TripLeg: TransportMode.TRANSIT, 12.67m, 600.0s, (0.0, 0.0) -> (2.0, 2.0)>>"
        )

    def test_summarised_properties(self):
        trip_leg1 = r5py.r5.trip_leg.TripLeg(
            transport_mode=r5py.TransportMode.TRANSIT,
            departure_time=datetime.datetime(2023, 4, 25, 15, 30),
            distance=12.34,
            travel_time=datetime.timedelta(minutes=10),
            wait_time=datetime.timedelta(minutes=2),
            route_id="56A",
            geometry=shapely.LineString([[0, 0], [1, 1], [2, 2]]),
        )
        trip_leg2 = r5py.r5.trip_leg.TripLeg(
            transport_mode=r5py.TransportMode.TRANSIT,
            departure_time=datetime.datetime(2023, 4, 25, 15, 30),
            distance=98.76,
            travel_time=datetime.timedelta(minutes=30),
            wait_time=datetime.timedelta(minutes=5),
            route_id="54A",
            geometry=shapely.LineString([[2, 2], [34, 34]]),
        )

        trip = trip_leg1 + trip_leg2

        assert trip.distance == pytest.approx(111.10)
        assert trip.geometry.wkt == "LINESTRING (0 0, 1 1, 2 2, 34 34)"
        assert trip.route_ids == ["56A", "54A"]
        assert trip.transport_modes == [
            r5py.TransportMode.TRANSIT,
            r5py.TransportMode.TRANSIT,
        ]
        assert trip.travel_time == datetime.timedelta(minutes=(30 + 10))
        assert trip.wait_time == datetime.timedelta(minutes=(2 + 5))

    def test_as_table(self):
        trip_leg = r5py.r5.trip_leg.TripLeg(
            transport_mode=r5py.TransportMode.TRANSIT,
            departure_time=datetime.datetime(2023, 4, 25, 15, 30),
            distance=12.67,
            travel_time=datetime.timedelta(minutes=10),
            wait_time=datetime.timedelta(minutes=2),
            feed="GTFS",
            agency_id="HSR",
            route_id="48A",
            start_stop_id="90210",
            end_stop_id="10802",
            geometry=shapely.LineString([[0, 0], [1, 1], [2, 2]]),
        )
        trip = r5py.r5.trip.Trip([trip_leg])

        assert trip.as_table() == [
            [
                0,
                r5py.TransportMode.TRANSIT,
                datetime.datetime(2023, 4, 25, 15, 30),
                12.67,
                datetime.timedelta(minutes=10),
                datetime.timedelta(minutes=2),
                "GTFS",
                "HSR",
                "48A",
                "90210",
                "10802",
                shapely.LineString([[0, 0], [1, 1], [2, 2]]),
            ]
        ]

    def test_distance_none(self):
        trip = sum(
            [
                r5py.r5.trip_leg.TripLeg(distance=None),
                r5py.r5.trip_leg.TripLeg(distance=None),
            ]
        )
        assert trip.distance is None



#!/usr/bin/env python3


"""Fixtures describing the expected routing results."""


import pytest

from .routing_results import ISOCHRONES_WALK


ISOCHRONES_WALK_BLAKE2B = "5e91181cfdff18a3ebcf289eba1955b61b27260fa40d1cd669bb2f754c31b23b59ecd74ee9d3a31f2d51c9de32aff0781ff5c948eb4ea220b622525d94e6a625"
ISOCHRONES_WALK_BLAKE2S = (
    "da098b9fcf440e7252b2b213160dd810392c03ee07c58ababfddf7d4c5d8d502"
)
ISOCHRONES_WALK_SHA256 = (
    "2a444c35ef676928d2a70914c64e34c569b10a6473ee6d61c74d9740351d80cc"
)


@pytest.fixture
def file_digest_test_file_as_pathlib_path():
    """Return the path of a test file as a pathlib.Path."""
    yield ISOCHRONES_WALK


@pytest.fixture
def file_digest_test_file_as_str(file_digest_test_file_as_pathlib_path):
    """Return the path of a test file as a str."""
    yield f"{file_digest_test_file_as_pathlib_path}"


@pytest.fixture()
def file_digest_sha256():
    """Return the expected SHA256 hash for the test file."""
    yield ISOCHRONES_WALK_SHA256


@pytest.fixture()
def file_digest_blake2b():
    """Return the expected BLAKE2B hash for the test file."""
    yield ISOCHRONES_WALK_BLAKE2B


@pytest.fixture()
def file_digest_blake2s():
    """Return the expected BLAKE2S hash for the test file."""
    yield ISOCHRONES_WALK_BLAKE2S



#!/usr/bin/env python3


"""Represent one leg of a trip, specifically transfers between public transport vehicles."""


from .direct_leg import DirectLeg


__all__ = ["TransferLeg"]


class TransferLeg(DirectLeg):
    """Represent one leg of a trip, specifically transfers between public transport vehicles."""



#!/usr/bin/env python3

import contextlib
import datetime

import geopandas
import geopandas.testing
import pandas
import pytest
import pytest_lazy_fixtures
import shapely

import r5py
import r5py.util.exceptions


@pytest.fixture(autouse=True, scope="class")
def _expectations(
    request,
    can_compute_detailed_route_geometries,
):
    if can_compute_detailed_route_geometries:
        expectations = contextlib.nullcontext()
    else:
        expectations = pytest.warns(
            RuntimeWarning,
            match=(
                "R5 has been compiled with `TransitLayer.SAVE_SHAPES = false` "
                "\\(the default\\). The geometries of public transport routes "
                "are inaccurate \\(straight lines between stops\\), and distances "
                "can not be computed."
            ),
        )
    request.cls._expectations = expectations


class TestDetailedItinerariesInputValidation:
    @pytest.mark.parametrize(
        [
            "origins",
            "expected_error",
        ],
        [
            (
                pytest_lazy_fixtures.lf("origins_invalid_no_id"),
                r5py.util.exceptions.NoIDColumnError,
            ),
            (
                pytest_lazy_fixtures.lf("origins_invalid_duplicate_ids"),
                r5py.util.exceptions.NonUniqueIDError,
            ),
        ],
    )
    def test_origins_invalid_data(
        self,
        transport_network,
        origins,
        departure_datetime,
        expected_error,
    ):
        with pytest.raises(expected_error):
            _ = r5py.DetailedItineraries(
                transport_network,
                origins=origins,
                departure=departure_datetime,
            )

    def test_origins_valid_data(
        self,
        transport_network,
        origins_valid_ids,
        departure_datetime,
    ):
        with self._expectations:
            _ = r5py.DetailedItineraries(
                transport_network,
                origins=origins_valid_ids,
                departure=departure_datetime,
            )

    @pytest.mark.parametrize(
        [
            "origins",
            "destinations",
            "expected_error",
        ],
        [
            (
                pytest_lazy_fixtures.lf("origins_invalid_no_id"),
                pytest_lazy_fixtures.lf("origins_invalid_no_id"),
                r5py.util.exceptions.NoIDColumnError,
            ),
            (
                pytest_lazy_fixtures.lf("origins_invalid_duplicate_ids"),
                pytest_lazy_fixtures.lf("origins_invalid_duplicate_ids"),
                r5py.util.exceptions.NonUniqueIDError,
            ),
            (
                pytest_lazy_fixtures.lf("origins_invalid_no_id"),
                pytest_lazy_fixtures.lf("origins_invalid_duplicate_ids"),
                r5py.util.exceptions.NoIDColumnError,
            ),
            (
                pytest_lazy_fixtures.lf("origins_invalid_duplicate_ids"),
                pytest_lazy_fixtures.lf("origins_invalid_no_id"),
                r5py.util.exceptions.NonUniqueIDError,
            ),
        ],
    )
    def test_origins_and_destinations_invalid_data(
        self,
        transport_network,
        origins,
        destinations,
        departure_datetime,
        expected_error,
    ):
        with pytest.raises(expected_error):
            _ = r5py.DetailedItineraries(
                transport_network,
                origins=origins,
                destinations=destinations,
                departure=departure_datetime,
            )

    @pytest.mark.parametrize(
        [
            "origins",
            "destinations",
        ],
        [
            (
                pytest_lazy_fixtures.lf("origins_valid_ids"),
                pytest_lazy_fixtures.lf("origins_valid_ids"),
            )
        ],
    )
    def test_origins_and_destinations_valid_data(
        self,
        transport_network,
        origins,
        destinations,
        departure_datetime,
    ):
        with self._expectations:
            _ = r5py.DetailedItineraries(
                transport_network,
                origins=origins,
                destinations=destinations,
                departure=departure_datetime,
            )

    def test_try_to_route_without_origins(
        self,
        transport_network,
        departure_datetime,
    ):
        with pytest.raises(ValueError, match="No routing origins defined"):
            _ = r5py.DetailedItineraries(
                transport_network,
                departure=departure_datetime,
            )

    def test_try_to_route_without_destinations(
        self,
        transport_network,
        population_grid_points,
        departure_datetime,
    ):
        with self._expectations:
            detailed_itineraries = r5py.DetailedItineraries(
                transport_network,
                origins=population_grid_points[0:3],
                departure=departure_datetime,
            )

        pandas.testing.assert_frame_equal(
            detailed_itineraries.origins,
            detailed_itineraries.destinations,
        )


class TestDetailedItineraries:
    def test_detailed_itineraries_initialization(
        self,
        transport_network,
        population_grid_points_first_three,
        departure_datetime,
    ):
        with self._expectations:
            detailed_itineraries = r5py.DetailedItineraries(
                transport_network,
                origins=population_grid_points_first_three,
                departure=departure_datetime,
                transport_modes=[r5py.TransportMode.TRANSIT, r5py.TransportMode.WALK],
            )
        assert isinstance(detailed_itineraries, r5py.DetailedItineraries)
        assert isinstance(detailed_itineraries, geopandas.GeoDataFrame)

    def test_detailed_itineraries_initialization_with_files(
        self,
        transport_network_files_tuple,
        population_grid_points_first_three,
        departure_datetime,
    ):
        with self._expectations:
            detailed_itineraries = r5py.DetailedItineraries(
                transport_network_files_tuple,
                origins=population_grid_points_first_three,
                departure=departure_datetime,
                transport_modes=[r5py.TransportMode.TRANSIT, r5py.TransportMode.WALK],
            )
        assert isinstance(detailed_itineraries, r5py.DetailedItineraries)
        assert isinstance(detailed_itineraries, geopandas.GeoDataFrame)

    @pytest.mark.parametrize(
        [
            "origins",
            "destinations",
            "force_all_to_all",
            "expected_all_to_all",
            "expected_od_pairs_len",
        ],
        [
            (
                pytest_lazy_fixtures.lf("population_grid_points_first_three"),
                None,
                False,
                True,
                9,
            ),
            (
                pytest_lazy_fixtures.lf("population_grid_points_first_three"),
                pytest_lazy_fixtures.lf("population_grid_points_second_three"),
                False,
                False,
                3,
            ),
            (
                pytest_lazy_fixtures.lf("population_grid_points_first_three"),
                pytest_lazy_fixtures.lf("population_grid_points_second_three"),
                True,
                True,
                9,
            ),
            (
                pytest_lazy_fixtures.lf("population_grid_points_first_three"),
                pytest_lazy_fixtures.lf("population_grid_points_four"),
                False,
                True,
                12,
            ),
        ],
    )
    def test_od_pairs_all_to_all(
        self,
        transport_network,
        origins,
        destinations,
        force_all_to_all,
        departure_datetime,
        expected_all_to_all,
        expected_od_pairs_len,
    ):
        detailed_itineraries = r5py.DetailedItineraries(
            transport_network,
            origins=origins,
            destinations=destinations,
            force_all_to_all=force_all_to_all,
            departure=departure_datetime,
            transport_modes=[r5py.TransportMode.WALK],
        )
        assert detailed_itineraries.all_to_all == expected_all_to_all
        assert len(detailed_itineraries.od_pairs) == expected_od_pairs_len

    def test_gtfs_date_range_warnings(
        self,
        transport_network,
        population_grid_points,
        origin_point,
        departure_datetime,
    ):
        with (
            pytest.warns(
                RuntimeWarning,
                match=(
                    "Departure time .* is outside of the time range covered by "
                    "currently loaded GTFS data sets."
                ),
            ),
            self._expectations,
        ):
            _ = r5py.DetailedItineraries(
                transport_network,
                origins=origin_point,
                destinations=population_grid_points[::5],
                departure=datetime.datetime(2021, 2, 22, 8, 30),  # not in GTFS data set
                transport_modes=[r5py.TransportMode.TRANSIT, r5py.TransportMode.WALK],
            )

    def test_gtfs_date_range_warnings_without_gtfs_file(
        self,
        transport_network_from_test_files_without_gtfs,
        population_grid_points,
        origin_point,
        departure_datetime,
    ):
        with (
            pytest.warns(
                RuntimeWarning,
                match=(
                    "Departure time .* is outside of the time range covered by "
                    "currently loaded GTFS data sets."
                ),
            ),
            self._expectations,
        ):
            _ = r5py.DetailedItineraries(
                transport_network_from_test_files_without_gtfs,
                origins=origin_point,
                destinations=population_grid_points[::5],
                departure=departure_datetime,
                transport_modes=[r5py.TransportMode.TRANSIT, r5py.TransportMode.WALK],
            )

    @pytest.mark.parametrize(
        [
            "snap_to_network",
            "expected_snap_to_network",
        ],
        [
            (
                True,
                True,
            ),
            (
                False,
                False,
            ),
        ],
    )
    def test_snap_to_network_parameter(
        self,
        transport_network,
        population_grid_points,
        departure_datetime,
        snap_to_network,
        expected_snap_to_network,
    ):
        with self._expectations:
            detailed_itineraries = r5py.DetailedItineraries(
                transport_network,
                population_grid_points[::5],
                departure=departure_datetime,
                snap_to_network=snap_to_network,
            )
        assert detailed_itineraries.snap_to_network == expected_snap_to_network

    @pytest.mark.parametrize(
        [
            "snap_to_network",
            "expected_detailed_itineraries",
        ],
        [
            (
                True,
                pytest_lazy_fixtures.lf("walking_details_snapped"),
            ),
            (
                False,
                pytest_lazy_fixtures.lf("walking_details_not_snapped"),
            ),
        ],
    )
    def test_snap_to_network(
        self,
        transport_network,
        population_grid_points,
        departure_datetime,
        snap_to_network,
        expected_detailed_itineraries,
    ):
        # subset to keep test comparison data sets small
        origins = population_grid_points[::5]
        detailed_itineraries = r5py.DetailedItineraries(
            transport_network,
            origins=origins,
            departure=departure_datetime,
            snap_to_network=snap_to_network,
            transport_modes=[r5py.TransportMode.WALK],
        )

        detailed_itineraries = (
            detailed_itineraries.groupby(["from_id", "to_id", "option"])
            .sum(["travel_time", "distance"])
            .reset_index()
        )

        detailed_itineraries = detailed_itineraries.set_index(
            ["from_id", "to_id"]
        ).sort_index()
        expected_detailed_itineraries = expected_detailed_itineraries.set_index(
            ["from_id", "to_id"]
        ).sort_index()

        pandas.testing.assert_frame_equal(
            detailed_itineraries, expected_detailed_itineraries
        )

    def test_snap_to_network_with_unsnappable_origins(
        self,
        transport_network,
        unsnappable_points,
        population_grid_points,
        departure_datetime,
    ):
        origins = pandas.concat(
            [population_grid_points[-3:], unsnappable_points]
        ).reset_index(drop=False)
        with pytest.warns(
            RuntimeWarning,
            match="Some (origin|destination) points could not be snapped to the street network",
        ):
            _ = r5py.DetailedItineraries(
                transport_network,
                origins,
                departure=departure_datetime,
                snap_to_network=True,
                transport_modes=[r5py.TransportMode.WALK],
            )

    def test_snap_to_network_with_only_unsnappable_origins(
        self,
        transport_network,
        unsnappable_points,
        departure_datetime,
    ):
        with pytest.raises(
            ValueError, match="After snapping, no valid origin points remain"
        ):
            with pytest.warns(
                RuntimeWarning,
                match="Some (origin|destination) points could not be snapped to the street network",
            ):
                _ = r5py.DetailedItineraries(
                    transport_network,
                    unsnappable_points,
                    departure=departure_datetime,
                    snap_to_network=True,
                    transport_modes=[r5py.TransportMode.WALK],
                )

    def test_snap_to_network_with_unsnappable_destinations(
        self,
        transport_network,
        population_grid_points,
        unsnappable_points,
        departure_datetime,
    ):
        destinations = pandas.concat(
            [population_grid_points[-3:], unsnappable_points]
        ).reset_index(drop=False)
        with pytest.warns(
            RuntimeWarning,
            match="Some destination points could not be snapped to the street network",
        ):
            _ = r5py.DetailedItineraries(
                transport_network,
                origins=population_grid_points,
                destinations=destinations,
                departure=departure_datetime,
                snap_to_network=True,
                transport_modes=[r5py.TransportMode.WALK],
            )

    def test_snap_to_network_with_only_unsnappable_destinations(
        self,
        transport_network,
        population_grid_points,
        unsnappable_points,
        departure_datetime,
    ):
        with pytest.raises(
            ValueError, match="After snapping, no valid destination points remain"
        ):
            with pytest.warns(
                RuntimeWarning,
                match="Some destination points could not be snapped to the street network",
            ):
                _ = r5py.DetailedItineraries(
                    transport_network,
                    origins=population_grid_points[::5],
                    destinations=unsnappable_points,
                    departure=departure_datetime,
                    snap_to_network=True,
                    transport_modes=[r5py.TransportMode.WALK],
                )

    @pytest.mark.parametrize("snap_to_network", [True, False])
    def test_travel_details_between_identical_from_and_to_ids(
        self,
        transport_network,
        population_grid_points,
        departure_datetime,
        snap_to_network,
    ):
        detailed_itineraries = r5py.DetailedItineraries(
            transport_network,
            origins=population_grid_points[::5],
            transport_modes=[r5py.TransportMode.WALK],
            departure=departure_datetime,
            snap_to_network=snap_to_network,
        )

        assert detailed_itineraries[
            detailed_itineraries["from_id"] == detailed_itineraries["to_id"]
        ].travel_time.max() == datetime.timedelta(seconds=0)

    @pytest.mark.parametrize(
        [
            "transport_mode",
            "expected_travel_details",
        ],
        [
            (
                r5py.TransportMode.BICYCLE,
                pytest_lazy_fixtures.lf("detailed_itineraries_bicycle"),
            ),
            (
                r5py.TransportMode.CAR,
                pytest_lazy_fixtures.lf("detailed_itineraries_car"),
            ),
            (
                r5py.TransportMode.TRANSIT,
                pytest_lazy_fixtures.lf("detailed_itineraries_transit"),
            ),
            (
                r5py.TransportMode.WALK,
                pytest_lazy_fixtures.lf("detailed_itineraries_walk"),
            ),
        ],
    )
    def test_travel_details_by_transport_mode(
        self,
        transport_network,
        population_grid_points,
        departure_datetime,
        transport_mode,
        expected_travel_details,
        can_compute_detailed_route_geometries,
    ):
        # subset to keep test comparison data sets small
        origins = population_grid_points[::5].copy()

        if transport_mode == r5py.TransportMode.TRANSIT:
            expectations = self._expectations
        else:
            expectations = contextlib.nullcontext()

        with expectations:
            travel_details = r5py.DetailedItineraries(
                transport_network,
                origins=origins,
                departure=departure_datetime,
                departure_time_window=datetime.timedelta(
                    hours=1
                ),  # using old default for simplicity
                transport_modes=[transport_mode],
            )

        travel_details.travel_time = travel_details.travel_time.apply(
            lambda t: t.total_seconds()
        )
        travel_details.wait_time = travel_details.wait_time.apply(
            lambda t: t.total_seconds()
        )
        travel_details.transport_mode = travel_details.transport_mode.apply(
            lambda t: t.value
        )
        travel_details["departure_time"] = travel_details["departure_time"].astype(
            "datetime64[ms]"
        )

        travel_details = geopandas.GeoDataFrame(travel_details, crs="EPSG:4326")

        travel_details["geometry"] = travel_details["geometry"].apply(
            lambda geometry: (
                geometry
                if isinstance(geometry, shapely.MultiLineString)
                else shapely.MultiLineString([geometry])
            )
        )

        if (
            transport_mode == r5py.TransportMode.TRANSIT
            and not can_compute_detailed_route_geometries
        ):
            travel_details["distance"] = 0
            travel_details["geometry"] = None
            expected_travel_details["distance"] = 0
            expected_travel_details["geometry"] = None

        geopandas.testing.assert_geodataframe_equal(
            travel_details,
            expected_travel_details,
            check_less_precise=True,  # geometries
            check_dtype=False,
            normalize=True,
        )


class TestDetailedItinerariesComputer:
    def test_detailed_itineraries_warning_and_legacy_interface(
        self,
        transport_network,
        population_grid_points,
        origin_point,
        departure_datetime,
    ):
        with self._expectations:
            detailed_itineraries_new = r5py.DetailedItineraries(
                transport_network,
                origins=origin_point,
                destinations=population_grid_points[::5],
                departure=departure_datetime,
                transport_modes=[r5py.TransportMode.TRANSIT, r5py.TransportMode.WALK],
            )

        with (
            pytest.warns(
                DeprecationWarning,
                match=(
                    "Use `DetailedItineraries` instead, "
                    "`DetailedItinerariesComputer will be deprecated in a "
                    "future release."
                ),
            ),
            self._expectations,
        ):
            detailed_itineraries_old = r5py.DetailedItinerariesComputer(
                transport_network,
                origins=origin_point,
                destinations=population_grid_points[::5],
                departure=departure_datetime,
                transport_modes=[r5py.TransportMode.TRANSIT, r5py.TransportMode.WALK],
            ).compute_travel_details()

        geopandas.testing.assert_geodataframe_equal(
            detailed_itineraries_new, detailed_itineraries_old
        )



#!/usr/bin/env python3


"""Test whether routing is (more or less) deterministic (https://github.com/r5py/r5py/issues/240)."""


import itertools

import pandas.testing
import pytest

import r5py


@pytest.fixture(scope="module")
def intermediate_results():
    # will be modified in each iteration,
    # used to compare results across computation runs
    yield {}


# Python 3.10+
def pairwise(iterable):
    """pairwise("ABCDEFG") --> AB BC CD DE EF FG ."""
    try:
        return itertools.pairwise(iterable)
    except AttributeError:  # Python<3.10
        a, b = itertools.tee(iterable)
        next(b, None)
        return zip(a, b)


class TestDeterministicBehaviour:
    @pytest.mark.parametrize(
        "transport_modes",
        [
            (r5py.TransportMode.WALK,),
            (r5py.TransportMode.BICYCLE,),
            (r5py.TransportMode.CAR,),
            (r5py.TransportMode.TRANSIT, r5py.TransportMode.WALK),
        ],
    )
    @pytest.mark.parametrize("iteration", range(5))
    def test_public_transport_travel_time_matrices(
        self,
        transport_network,
        population_grid_points,
        departure_datetime,
        transport_modes,
        intermediate_results,
        iteration,
    ):
        # subset to keep test comparison data sets small
        origins = population_grid_points[::5].copy()

        travel_times = r5py.TravelTimeMatrix(
            transport_network,
            origins,
            departure=departure_datetime,
            transport_modes=transport_modes,
            snap_to_network=True,
        )

        if transport_modes in intermediate_results:
            previous_results = intermediate_results[transport_modes]
            pandas.testing.assert_frame_equal(
                travel_times,
                previous_results,
                check_like=True,
            )
        intermediate_results[transport_modes] = travel_times.copy()



#!/usr/bin/env python3


"""Wraps a com.conveyal.r5.transit.TransitLayer."""


import datetime
import functools

import jpype
import jpype.types

from ..util import parse_int_date


__all__ = ["TransitLayer"]


class TransitLayer:
    """Wrap a com.conveyal.r5.transit.TransitLayer."""

    @classmethod
    def from_r5_transit_layer(cls, transit_layer):
        """
        Create a TransitLayer from a com.conveyal.r5.transit.TransitLayer.

        Arguments
        ---------
        transit_layer : com.conveyal.r5.transit.TransitLayer
        """
        instance = cls()
        instance._transit_layer = transit_layer
        return instance

    @functools.cached_property
    def start_date(self):
        """The earliest date the loaded GTFS data covers."""
        try:
            start_date = min(
                [
                    parse_int_date(service.calendar.start_date)
                    for service in self._transit_layer.services
                ]
            )
        except (AttributeError, ValueError) as exception:
            raise ValueError("No GTFS data set loaded") from exception
        return start_date

    @functools.cached_property
    def end_date(self):
        """The latest date the loaded GTFS data covers."""
        try:
            end_date = max(
                [
                    parse_int_date(service.calendar.end_date)
                    for service in self._transit_layer.services
                ]
            )
            end_date += datetime.timedelta(
                hours=23, minutes=59, seconds=59
            )  # *end* of day
        except (AttributeError, ValueError) as exception:
            raise ValueError("No GTFS data set loaded") from exception
        return end_date

    def covers(self, point_in_time):
        """Check whether `point_in_time` is covered by GTFS data sets."""
        try:
            covers = self.start_date <= point_in_time <= self.end_date
        except ValueError:  # no GTFS data loaded
            covers = False
        return covers

    def get_street_vertex_for_stop(self, stop):
        """
        Get the street layer’s vertex corresponding to `stop`.

        Arguments
        ---------
        stop : int
            ID of the public transport stop for which to find a vertex

        Returns
        -------
        int
            ID of the vertex corresponding to the public transport stop
        """
        street_vertex = self._transit_layer.streetVertexForStop.get(stop)
        return street_vertex

    @functools.cached_property
    def routes(self):
        """Return a list of GTFS routes."""
        return list(self._transit_layer.routes)

    @functools.cached_property
    def trip_patterns(self):
        """Return a list of GTFS trip patterns."""
        return list(self._transit_layer.tripPatterns)

    def get_stop_id_from_index(self, stop_index):
        """Get the GTFS stop id for the `stop_index`-th stop of this transit layer."""
        return self._transit_layer.stopIdForIndex[stop_index]


@jpype._jcustomizer.JConversion(
    "com.conveyal.r5.transit.TransitLayer", exact=TransitLayer
)
def _cast_TransitLayer(java_class, object_):
    return object_._transit_layer



#!/usr/bin/env python3


import pytest

import r5py.util


class TestSnakeToCamelCase:
    @pytest.mark.parametrize(
        ["snake_case", "camel_case"],
        [
            ("camel_case", "camelCase"),
            ("get_http_response_code", "getHttpResponseCode"),
            ("http_response_code_xyz", "httpResponseCodeXyz"),
            ("snake", "snake"),
        ],
    )
    def test_case_snake_case(self, snake_case, camel_case):
        assert r5py.util.snake_to_camel_case(snake_case) == camel_case



#!/usr/bin/env python3


"""Fixtures to be used in r5py tests."""


from .destinations import (
    population_grid,
    population_grid_points,
    population_grid_points_first_three,
    population_grid_points_second_three,
    population_grid_points_four,
    snapped_population_grid_points,
    unreachable_stops,
    unsnappable_points,
)

from .file_digest import (
    file_digest_test_file_as_pathlib_path,
    file_digest_test_file_as_str,
    file_digest_sha256,
    file_digest_blake2b,
    file_digest_blake2s,
)

from .garbage_collection import java_garbage_collection

from .origins import (
    multiple_origins,
    origin_point,
    origins_invalid_duplicate_ids,
    origins_invalid_no_id,
    origins_valid_ids,
)

from .r5_jar import (
    r5_jar_cached,
    r5_jar_cached_invalid,
    r5_jar_sha256,
    r5_jar_sha256_github_error_message_when_posting,
    r5_jar_sha256_invalid,
    r5_jar_url,
)

from .routing_parameters import (
    departure_datetime,
    regional_task,
)

from .routing_results import (
    detailed_itineraries_bicycle,
    detailed_itineraries_car,
    detailed_itineraries_transit,
    detailed_itineraries_walk,
    isochrones_from_multiple_origins,
    isochrones_bicycle,
    isochrones_car,
    isochrones_transit,
    isochrones_walk,
    travel_times_bicycle,
    travel_times_car,
    travel_times_transit,
    travel_times_walk,
    walking_details_not_snapped,
    walking_details_snapped,
    walking_times_not_snapped,
    walking_times_snapped,
)

from .sample_data import (
    sample_data_set_sha256,
    sample_data_set_url,
)

from .transport_network import (
    gtfs_file_path,
    gtfs_timezone_helsinki,
    not_a_gtfs_file,
    osm_pbf_file_path,
    transport_network,
    transport_network_files_tuple,
    transport_network_from_test_directory,
    transport_network_from_test_files,
    transport_network_from_test_files_without_gtfs,
)

from .upstream_r5 import (
    can_compute_detailed_route_geometries,
)

__all__ = [
    "can_compute_detailed_route_geometries",
    "departure_datetime",
    "detailed_itineraries_bicycle",
    "detailed_itineraries_car",
    "detailed_itineraries_transit",
    "detailed_itineraries_walk",
    "file_digest_test_file_as_pathlib_path",
    "file_digest_test_file_as_str",
    "file_digest_sha256",
    "file_digest_blake2b",
    "file_digest_blake2s",
    "java_garbage_collection",
    "gtfs_file_path",
    "gtfs_timezone_helsinki",
    "isochrones_from_multiple_origins",
    "isochrones_bicycle",
    "isochrones_car",
    "isochrones_transit",
    "isochrones_walk",
    "multiple_origins",
    "not_a_gtfs_file",
    "origin_point",
    "origins_invalid_duplicate_ids",
    "origins_invalid_no_id",
    "origins_valid_ids",
    "osm_pbf_file_path",
    "population_grid",
    "population_grid_points",
    "population_grid_points_first_three",
    "population_grid_points_four",
    "population_grid_points_second_three",
    "regional_task",
    "r5_jar_cached",
    "r5_jar_cached_invalid",
    "r5_jar_sha256",
    "r5_jar_sha256_github_error_message_when_posting",
    "r5_jar_sha256_invalid",
    "r5_jar_url",
    "sample_data_set_sha256",
    "sample_data_set_url",
    "snapped_population_grid_points",
    "transport_network",
    "transport_network_files_tuple",
    "transport_network_from_test_directory",
    "transport_network_from_test_files",
    "transport_network_from_test_files_without_gtfs",
    "travel_times_bicycle",
    "travel_times_car",
    "travel_times_transit",
    "travel_times_walk",
    "unreachable_stops",
    "unsnappable_points",
    "walking_details_not_snapped",
    "walking_details_snapped",
    "walking_times_not_snapped",
    "walking_times_snapped",
]



#!/usr/bin/env python3

"""Parse an integer date value (as return by com.conveyal.gtfs.GtfsFeed.services.end_date)."""


import datetime


__all__ = ["parse_int_date"]


def parse_int_date(int_date):
    """
    Parse an integer date value.

    com.conveyal.gtfs.GtfsFeed.services.start_date and .end_date are integer
    values in the format YYYYMMDD, so, e.g., 20220222 for 2 February 2022.
    This function converts such an integer value into the respective
    datetime.datetime (at 0:00 o’clock).

    """
    year = round(int_date / 10000)
    month = round((int_date % 10000) / 100)
    day = int_date % 100

    return datetime.datetime(year, month, day)



#!/usr/bin/env python3


import pytest

import r5py
import com.conveyal.r5


class TestTransportMode:
    @pytest.mark.parametrize(
        ["enum_member", "enum_name", "java_type"],
        [
            (r5py.TransportMode.AIR, "AIR", com.conveyal.r5.api.util.TransitModes),
            (r5py.TransportMode.BUS, "BUS", com.conveyal.r5.api.util.TransitModes),
            (
                r5py.TransportMode.CABLE_CAR,
                "CABLE_CAR",
                com.conveyal.r5.api.util.TransitModes,
            ),
            (r5py.TransportMode.FERRY, "FERRY", com.conveyal.r5.api.util.TransitModes),
            (
                r5py.TransportMode.FUNICULAR,
                "FUNICULAR",
                com.conveyal.r5.api.util.TransitModes,
            ),
            (r5py.TransportMode.RAIL, "RAIL", com.conveyal.r5.api.util.TransitModes),
            (
                r5py.TransportMode.SUBWAY,
                "SUBWAY",
                com.conveyal.r5.api.util.TransitModes,
            ),
            (r5py.TransportMode.TRAM, "TRAM", com.conveyal.r5.api.util.TransitModes),
            (
                r5py.TransportMode.TRANSIT,
                "TRANSIT",
                com.conveyal.r5.api.util.TransitModes,
            ),
            (r5py.TransportMode.BICYCLE, "BICYCLE", com.conveyal.r5.profile.StreetMode),
            (r5py.TransportMode.CAR, "CAR", com.conveyal.r5.profile.StreetMode),
            (r5py.TransportMode.WALK, "WALK", com.conveyal.r5.profile.StreetMode),
            (r5py.TransportMode.BICYCLE, "BICYCLE", com.conveyal.r5.api.util.LegMode),
            (r5py.TransportMode.CAR, "CAR", com.conveyal.r5.api.util.LegMode),
            (r5py.TransportMode.WALK, "WALK", com.conveyal.r5.api.util.LegMode),
            (
                r5py.TransportMode.BICYCLE_RENT,
                "BICYCLE_RENT",
                com.conveyal.r5.api.util.LegMode,
            ),
            (r5py.TransportMode.CAR_PARK, "CAR_PARK", com.conveyal.r5.api.util.LegMode),
        ],
    )
    def test_transportmode(self, enum_member, enum_name, java_type):
        assert enum_member.name == enum_name
        assert enum_member.name == java_type.valueOf(enum_name).name()

    @pytest.mark.parametrize(
        ["str_transport_mode", "expected_transport_mode"],
        [
            ("AIR", r5py.TransportMode.AIR),
            ("BUS", r5py.TransportMode.BUS),
            ("CABLE_CAR", r5py.TransportMode.CABLE_CAR),
            ("FERRY", r5py.TransportMode.FERRY),
            ("FUNICULAR", r5py.TransportMode.FUNICULAR),
            ("GONDOLA", r5py.TransportMode.GONDOLA),
            ("RAIL", r5py.TransportMode.RAIL),
            ("SUBWAY", r5py.TransportMode.SUBWAY),
            ("TRAM", r5py.TransportMode.TRAM),
            ("TRANSIT", r5py.TransportMode.TRANSIT),
            ("BICYCLE", r5py.TransportMode.BICYCLE),
            ("CAR", r5py.TransportMode.CAR),
            ("WALK", r5py.TransportMode.WALK),
            ("BICYCLE_RENT", r5py.TransportMode.BICYCLE_RENT),
            ("CAR_PARK", r5py.TransportMode.CAR_PARK),
            ("air", r5py.TransportMode.AIR),
            ("bus", r5py.TransportMode.BUS),
            ("cable_car", r5py.TransportMode.CABLE_CAR),
            ("ferry", r5py.TransportMode.FERRY),
            ("funicular", r5py.TransportMode.FUNICULAR),
            ("gondola", r5py.TransportMode.GONDOLA),
            ("rail", r5py.TransportMode.RAIL),
            ("subway", r5py.TransportMode.SUBWAY),
            ("tram", r5py.TransportMode.TRAM),
            ("transit", r5py.TransportMode.TRANSIT),
            ("bicycle", r5py.TransportMode.BICYCLE),
            ("car", r5py.TransportMode.CAR),
            ("walk", r5py.TransportMode.WALK),
            ("bicycle_rent", r5py.TransportMode.BICYCLE_RENT),
            ("car_park", r5py.TransportMode.CAR_PARK),
            ("AiR", r5py.TransportMode.AIR),
            ("bUs", r5py.TransportMode.BUS),
            ("CaBlE_CaR", r5py.TransportMode.CABLE_CAR),
            ("fErRy", r5py.TransportMode.FERRY),
            ("FuNiCuLaR", r5py.TransportMode.FUNICULAR),
            ("gOnDoLa", r5py.TransportMode.GONDOLA),
            ("RaIl", r5py.TransportMode.RAIL),
            ("SuBwAy", r5py.TransportMode.SUBWAY),
            ("TrAm", r5py.TransportMode.TRAM),
            ("TrAnSiT", r5py.TransportMode.TRANSIT),
            ("bIcYcLe", r5py.TransportMode.BICYCLE),
            ("CaR", r5py.TransportMode.CAR),
            ("wAlK", r5py.TransportMode.WALK),
            ("bIcYcLe_ReNt", r5py.TransportMode.BICYCLE_RENT),
            ("CaR_PaRk", r5py.TransportMode.CAR_PARK),
        ],
    )
    def test_str_transportmode(self, str_transport_mode, expected_transport_mode):
        assert r5py.TransportMode(str_transport_mode) == expected_transport_mode

    @pytest.mark.parametrize(
        ["invalid_transport_mode"],
        [
            ("Helicopter",),
            ("adsffoobar",),
            (13,),
            (None,),
            ("1234",),
        ],
    )
    def test_invalid_transportmode(self, invalid_transport_mode):
        with pytest.raises(ValueError, match="is not a valid TransportMode"):
            _ = r5py.TransportMode(invalid_transport_mode)

    @pytest.mark.parametrize(
        ["mode", "expected"],
        [
            ("CAR", True),
            ("BICYCLE", True),
            (r5py.TransportMode.WALK, True),
            (r5py.TransportMode.BICYCLE_RENT, True),
            (r5py.TransportMode.BUS, False),
            ("TRANSIT", False),
        ],
    )
    def test_is_leg_mode(self, mode, expected):
        assert r5py.TransportMode(mode).is_leg_mode == expected

    @pytest.mark.parametrize(
        ["mode", "expected"],
        [
            ("CAR", True),
            ("BICYCLE", True),
            (r5py.TransportMode.WALK, True),
            (r5py.TransportMode.BICYCLE_RENT, False),
            (r5py.TransportMode.BUS, False),
            ("TRANSIT", False),
        ],
    )
    def test_is_street_mode(self, mode, expected):
        assert r5py.TransportMode(mode).is_street_mode == expected

    @pytest.mark.parametrize(
        ["mode", "expected"],
        [
            ("CAR", False),
            ("BICYCLE", False),
            (r5py.TransportMode.WALK, False),
            (r5py.TransportMode.BICYCLE_RENT, False),
            (r5py.TransportMode.BUS, True),
            ("TRANSIT", True),
        ],
    )
    def test_is_transit_mode(self, mode, expected):
        assert r5py.TransportMode(mode).is_transit_mode == expected

    def test_add(self):
        transport_mode1 = r5py.TransportMode.TRANSIT
        transport_mode2 = r5py.TransportMode.WALK

        transport_modes = transport_mode1 + transport_mode2
        assert isinstance(transport_modes, list)
        assert transport_modes == [transport_mode1, transport_mode2]

        transport_modes = transport_mode2 + transport_mode1
        assert isinstance(transport_modes, list)
        assert transport_modes == [transport_mode2, transport_mode1]

        transport_modes = sum([transport_mode1, transport_mode2])
        assert isinstance(transport_modes, list)
        assert transport_modes == [transport_mode1, transport_mode2]

        transport_modes = [transport_mode1] + transport_mode2
        assert isinstance(transport_modes, list)
        assert transport_modes == [transport_mode1, transport_mode2]

        transport_modes = transport_mode1 + [transport_mode2]
        assert isinstance(transport_modes, list)
        assert transport_modes == [transport_mode1, transport_mode2]

    @pytest.mark.parametrize(
        ["invalid_other"],
        [
            (123.0,),
            (1,),
            ("asdfasdf",),
            ({},),
        ],
    )
    def test_add_invalid_type(self, invalid_other):
        transport_mode = r5py.TransportMode.TRANSIT
        with pytest.raises(TypeError, match="unsupported operand type"):
            _ = transport_mode + invalid_other

        with pytest.raises(TypeError, match="unsupported operand type"):
            _ = invalid_other + transport_mode



#!/usr/bin/env python3


import filecmp
import pathlib
import random

from .temporary_directory import TemporaryDirectory

import r5py.util


class TestWorkingDirectory:
    def test_working_directory(self):
        with TemporaryDirectory() as temp_directory:
            input_file = pathlib.Path(temp_directory) / "test_input.txt"
            input_file.write_text(f"{random.randrange(16**5):032x}")

            working_copy = r5py.util.WorkingCopy(input_file)

            assert filecmp.cmp(input_file, working_copy, shallow=False)

            working_copy.unlink()

    def test_working_directory_failed_symlink(self, monkeypatch):
        def _symlink_to(*args, **kwargs):
            raise OSError

        monkeypatch.setattr(pathlib.Path, "symlink_to", _symlink_to)

        with TemporaryDirectory() as temp_directory:
            input_file = pathlib.Path(temp_directory) / "test_input.txt"
            input_file.write_text(f"{random.randrange(16**5):032x}")

            working_copy = r5py.util.WorkingCopy(input_file)

            assert filecmp.cmp(input_file, working_copy, shallow=False)

            working_copy.unlink()



#!/usr/bin/env python3

"""A less complex representation of com.conveyal.r5.api.util.StreetSegment."""


import datetime

import shapely


__all__ = ["StreetSegment"]


class StreetSegment:
    """A less complex representation of com.conveyal.r5.api.util.StreetSegment."""

    distance = 0
    duration = datetime.timedelta()
    geometry = shapely.LineString()

    def __init__(self, street_path):
        """
        Initialise a less complex representation of com.conveyal.r5.api.util.StreetSegment.

        Arguments
        ---------
        street_path : com.conveyal.r5.profile.StreetPath
            StreetPath, obtained, e.g., from StreetRouter state
        """
        self.distance = street_path.getDistance()
        self.duration = street_path.getDuration()
        self.geometry = shapely.line_merge(
            shapely.MultiLineString(
                [
                    shapely.from_wkt(
                        str(street_path.getEdge(edge).getGeometry().toText())
                    )
                    for edge in street_path.getEdges()
                ]
            )
        )



#!/usr/bin/env python3


"""The transport modes supported by R5 (Leg, Street, Transit, combined)."""


import enum

import jpype

from ..util import start_jvm

import com.conveyal.r5


__all__ = ["TransportMode"]


start_jvm()


TRANSIT_MODES = [
    "AIR",
    "BUS",
    "CABLE_CAR",
    "FERRY",
    "FUNICULAR",
    "GONDOLA",
    "RAIL",
    "SUBWAY",
    "TRAM",
    "TRANSIT",
]

STREET_MODES = [
    "BICYCLE",
    "CAR",
    "WALK",
]

LEG_MODES = STREET_MODES + [
    "BICYCLE_RENT",
    "CAR_PARK",
]


class TransportMode(enum.Enum):
    """
    Transport modes.

    TransportMode.AIR, TransportMode.TRAM, TransportMode.SUBWAY,
    TransportMode.RAIL, TransportMode.BUS, TransportMode.FERRY,
    TransportMode.CABLE_CAR, TransportMode.GONDOLA, TransportMode.FUNICULAR,
    TransportMode.TRANSIT (translate into R5’s TransitMode)

    TransportMode.WALK, TransportMode.BICYCLE, TransportMode.CAR (translate into
    R5’s StreetMode or LegMode)

    TransportMode.BICYCLE_RENT, TransportMode.CAR_PARK (translate into R5’s LegMode)
    """

    @classmethod
    def _missing_(cls, value):
        value = str(value).upper()
        for member in cls:
            if value == member.value:
                return member
        return None

    def __add__(self, other):
        """Combine two transport modes."""
        if isinstance(other, self.__class__):
            return [self, other]
        elif isinstance(other, list):
            return [self] + other
        else:
            raise TypeError(
                f"unsupported operand type(s) for '+': '{type(other)}' and '{type(self)}'"
            )

    def __radd__(self, other):
        """Combine two transport modes."""
        if other == 0:  # first iteration of sum()
            return self
        elif isinstance(other, list):
            return other + [self]
        else:
            return self.__add__(other)

    @property
    def is_leg_mode(self):
        """Can this TransportMode function as a LegMode?."""
        return self.name in LEG_MODES

    @property
    def is_street_mode(self):
        """Can this TransportMode function as a StreetMode?."""
        return self.name in STREET_MODES

    @property
    def is_transit_mode(self):
        """Can this TransportMode function as a TransitMode?."""
        return self.name in TRANSIT_MODES

    AIR = "AIR"
    BUS = "BUS"
    CABLE_CAR = "CABLE_CAR"
    FERRY = "FERRY"
    FUNICULAR = "FUNICULAR"
    GONDOLA = "GONDOLA"
    RAIL = "RAIL"
    SUBWAY = "SUBWAY"
    TRAM = "TRAM"
    TRANSIT = "TRANSIT"

    BICYCLE = "BICYCLE"
    CAR = "CAR"
    WALK = "WALK"

    BICYCLE_RENT = "BICYCLE_RENT"
    CAR_PARK = "CAR_PARK"


@jpype._jcustomizer.JConversion("com.conveyal.r5.api.util.LegMode", exact=TransportMode)
def _cast_LegMode(java_class, object_):
    if object_.name in LEG_MODES:
        return com.conveyal.r5.api.util.LegMode.valueOf(object_.name)
    else:
        raise ValueError(f"{object_.name} is not a valid R5 LegMode")


@jpype._jcustomizer.JConversion(
    "com.conveyal.r5.profile.StreetMode", exact=TransportMode
)
def _cast_StreetMode(java_class, object_):
    if object_.name in STREET_MODES:
        return com.conveyal.r5.profile.StreetMode.valueOf(object_.name)
    else:
        raise ValueError(f"{object_.name} is not a valid R5 StreetMode")


@jpype._jcustomizer.JConversion(
    "com.conveyal.r5.api.util.TransitModes", exact=TransportMode
)
def _cast_TransitMode(java_class, object_):
    if object_.name in TRANSIT_MODES:
        return com.conveyal.r5.api.util.TransitModes.valueOf(object_.name)
    else:
        raise ValueError(f"{object_.name} is not a valid R5 TransitModes")



#!/usr/bin/env python3

"""R5 classes."""

from .access_leg import AccessLeg
from .detailed_itineraries import DetailedItineraries, DetailedItinerariesComputer
from .direct_leg import DirectLeg
from .egress_leg import EgressLeg
from .isochrones import Isochrones
from .regional_task import RegionalTask
from .scenario import Scenario
from .street_layer import StreetLayer
from .transfer_leg import TransferLeg
from .transit_leg import TransitLeg
from .transport_mode import TransportMode
from .transport_network import TransportNetwork
from .travel_time_matrix import TravelTimeMatrix, TravelTimeMatrixComputer
from .trip import Trip
from .trip_planner import TripPlanner

__all__ = [
    "AccessLeg",
    "DetailedItineraries",
    "DetailedItinerariesComputer",
    "DirectLeg",
    "EgressLeg",
    "Isochrones",
    "RegionalTask",
    "Scenario",
    "SpeedConfig",
    "StreetLayer",
    "TransferLeg",
    "TransitLeg",
    "TransportMode",
    "TransportNetwork",
    "TravelTimeMatrix",
    "TravelTimeMatrixComputer",
    "Trip",
    "TripPlanner",
]



#!/usr/bin/env python3


"""Fixtures related to testing the SampleData class."""


import pytest


SAMPLE_DATA_SET_URL = "https://raw.githubusercontent.com/r5py/r5py.sampledata.sao_paulo/main/data/spo_hexgrid.csv"
SAMPLE_DATA_SET_SHA256 = (
    "769660f8f1bc95d2741bbc4225e5e0e77e73461ea8b3e225a58e397b0748bdd4"
)


@pytest.fixture
def sample_data_set_sha256():
    """Return the SHA256 hash of the sample data at `sample_data_set_url()`."""
    yield SAMPLE_DATA_SET_SHA256


@pytest.fixture
def sample_data_set_url():
    """Return the web address from which a sample data set can be downloaded."""
    yield SAMPLE_DATA_SET_URL



#!/usr/bin/env python3


import datetime

import pytest

import r5py.util


class TestParseIntDate:
    @pytest.mark.parametrize(
        ["int_date", "expected"],
        [
            (20220202, datetime.datetime(2022, 2, 2)),
            (19990202, datetime.datetime(1999, 2, 2)),
            (20191231, datetime.datetime(2019, 12, 31)),
        ],
    )
    def test_parse_int_date(self, int_date, expected):
        assert r5py.util.parse_int_date(int_date) == expected



#!/usr/bin/env python3


import pytest
import pytest_lazy_fixtures

import r5py


class TestContainsGtfsData:
    @pytest.mark.parametrize(
        ["path", "expected"],
        [
            (pytest_lazy_fixtures.lf("gtfs_file_path"), True),
            (pytest_lazy_fixtures.lf("not_a_gtfs_file"), False),
        ],
    )
    def test_contains_gtfs_data(self, path, expected):
        assert r5py.util.contains_gtfs_data(path) == expected
        assert r5py.util.contains_gtfs_data(str(path)) == expected
        with open(path, "rb") as f:
            r5py.util.contains_gtfs_data(f)



#!/usr/bin/env python3

"""Check whether a file contains a GTFS data set."""


import zipfile


__all__ = ["contains_gtfs_data"]


# https://developers.google.com/transit/gtfs/reference#field_definitions
GTFS_REQUIRED_FILES = (
    "agency.txt",
    "stops.txt",
    "routes.txt",
    "trips.txt",
    "stop_times.txt",
)


def contains_gtfs_data(path):
    """
    Check whether the file in `path` contains a GTFS dataset.

    This is a rather heuristic approach: it tries to open the file
    as a ZIP archive, and confirm the presence of the files required
    by the GTFS standard reference.

    Arguments:
    ----------
    path : str | pathlib.Path | file-like
        The file to check. Should be opened in binary mode, if passed as a
        file-like object.

    Returns:
    --------
    bool : `True` if `path` likely contains a GTFS dataset, `False` if not.
    """
    try:
        archive = zipfile.ZipFile(path)
        assert all(
            gtfs_field in archive.namelist() for gtfs_field in GTFS_REQUIRED_FILES
        )
        contains_gtfs_data = True
    except (AssertionError, FileNotFoundError, zipfile.BadZipFile):
        contains_gtfs_data = False
    return contains_gtfs_data



#!/usr/bin/env python3

"""Convert a camelCase/CamelCase formated string to a snake_case format."""


import re


__all__ = ["camel_to_snake_case"]


# https://stackoverflow.com/a/1176023


CAMEL_CASE_TO_SNAKE_CASE_RE1 = re.compile("(.)([A-Z][a-z]+)")
CAMEL_CASE_TO_SNAKE_CASE_RE2 = re.compile("([a-z0-9])([A-Z])")
CAMEL_CASE_TO_SNAKE_CASE_SUBSTITUTE = r"\1_\2"


def camel_to_snake_case(camel_case):
    """Convert `camel_case` to snake_case spelling."""
    return CAMEL_CASE_TO_SNAKE_CASE_RE2.sub(
        CAMEL_CASE_TO_SNAKE_CASE_SUBSTITUTE,
        CAMEL_CASE_TO_SNAKE_CASE_RE1.sub(
            CAMEL_CASE_TO_SNAKE_CASE_SUBSTITUTE, camel_case
        ),
    ).lower()



#!/usr/bin/env python3


"""Compute polygons of equal travel time from a destination."""


import datetime
import warnings

import geohexgrid
import geopandas
import pandas
import pyproj
import shapely
import simplification.cutil

from .base_travel_time_matrix import BaseTravelTimeMatrix
from .transport_mode import TransportMode
from .travel_time_matrix import TravelTimeMatrix
from ..util import GoodEnoughEquidistantCrs, SpatiallyClusteredGeoDataFrame


__all__ = ["Isochrones"]


EMPTY_POINT = shapely.Point()
R5_CRS = "EPSG:4326"

CONCAVE_HULL_BUFFER_SIZE = 20.0  # metres
CONCAVE_HULL_RATIO = 0.3

VERY_SMALL_BUFFER_SIZE = 0.001  # turn points into polygons


class Isochrones(BaseTravelTimeMatrix):
    """Compute polygons of equal travel time from a destination."""

    _r5py_attributes = BaseTravelTimeMatrix._r5py_attributes + [
        "_isochrones",
        "isochrones",
        "point_grid_resolution",
        "point_grid_sample_ratio",
    ]

    def __init__(
        self,
        transport_network,
        origins,
        isochrones=pandas.timedelta_range(
            start=datetime.timedelta(minutes=0),
            end=datetime.timedelta(hours=1),
            freq=datetime.timedelta(minutes=15),
        ),
        point_grid_resolution=100,
        point_grid_sample_ratio=1.0,
        **kwargs,
    ):
        """
        Compute polygons of equal travel time from one or more destinations.

        ``r5py.Isochrones`` are child classes of ``geopandas.GeoDataFrame`` and
        support all of their methods and properties, see
        https://geopandas.org/en/stable/docs.html

        Arguments
        ---------
        transport_network : r5py.TransportNetwork | tuple(str, list(str), dict)
            The transport network to route on. This can either be a readily
            initialised r5py.TransportNetwork or a tuple of the parameters
            passed to ``TransportNetwork.__init__()``: the path to an OpenStreetMap
            extract in PBF format, a list of zero of more paths to GTFS transport
            schedule files, and a dict with ``build_config`` options.
        origins : geopandas.GeoDataFrame | shapely.Point
            Place(s) to find a route _from_
            Must be/have a point geometry. If multiple origin points are passed,
            isochrones will be computed as minimum travel time from any of them.
        isochrones : pandas.TimedeltaIndex | collections.abc.Iterable[int]
            For which interval to compute isochrone polygons. An iterable of
            integers is interpreted as minutes.
        point_grid_resolution : int
            Distance in meters between points in the regular grid of points laid over the
            transport network’s extent that is used to compute isochrones.
            Increase this value for performance, decrease it for precision.
        point_grid_sample_ratio : float
            Share of points of the point grid that are used in computation,
            ranging from 0.01 to 1.0.
            Increase this value for performance, decrease it for precision.
        **kwargs : mixed
            Any arguments than can be passed to r5py.RegionalTask:
            ``departure``, ``departure_time_window``, ``percentiles``, ``transport_modes``,
            ``access_modes``, ``egress_modes``, ``max_time``, ``max_time_walking``,
            ``max_time_cycling``, ``max_time_driving``, ``speed_cycling``, ``speed_walking``,
            ``max_public_transport_rides``, ``max_bicycle_traffic_stress``
            Note that not all arguments might make sense in this context, and the
            underlying R5 engine might ignore some of them.
            If percentiles are specified, the lowest one will be used for
            isochrone computation.
        """
        geopandas.GeoDataFrame.__init__(self)
        BaseTravelTimeMatrix.__init__(
            self,
            transport_network,
            **kwargs,
        )

        self.EQUIDISTANT_CRS = GoodEnoughEquidistantCrs(self.transport_network.extent)

        if isinstance(origins, shapely.Geometry):
            origins = geopandas.GeoDataFrame(
                {
                    "id": [
                        "origin",
                    ],
                    "geometry": [
                        origins,
                    ],
                },
                crs=R5_CRS,
            )
        self.origins = origins
        self.isochrones = isochrones

        self.point_grid_resolution = point_grid_resolution
        self.point_grid_sample_ratio = max(0.01, min(1.0, point_grid_sample_ratio))

        travel_times = TravelTimeMatrix(
            transport_network,
            origins=self.origins,
            destinations=self.destinations,
            max_time=self.isochrones.max(),
            **kwargs,
        )

        data = self._compute_isochrones_from_travel_times(travel_times)

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=(
                    "You are adding a column named 'geometry' to a GeoDataFrame "
                    "constructed without an active geometry column"
                ),
                category=FutureWarning,
            )
            for column in data.columns:
                self[column] = data[column]
            self.set_geometry("geometry")

        del self.transport_network

    def _compute_isochrones_from_travel_times(self, travel_times):
        travel_times = travel_times.dropna().groupby("to_id").min().reset_index()

        if self.request.percentiles == [50]:
            travel_time_column = "travel_time"
        else:
            travel_time_column = f"travel_time_p{self.request.percentiles[0]:d}"

        isochrones = {
            "travel_time": [],
            "geometry": [],
        }

        for isochrone in self.isochrones:
            reached_nodes = (
                self.destinations.set_index("id")
                .join(
                    travel_times[
                        travel_times[travel_time_column]
                        <= (isochrone.total_seconds() / 60)
                    ].set_index("to_id"),
                    how="inner",
                )
                .reset_index()
            )

            # isochrone polygons might be disjoint (e.g., around metro stops)
            if not reached_nodes.empty:
                reached_nodes = SpatiallyClusteredGeoDataFrame(
                    reached_nodes, eps=(2.0 * self.point_grid_resolution)
                ).to_crs(self.EQUIDISTANT_CRS)
                isochrone_polygons = pandas.concat(
                    [
                        (
                            reached_nodes[reached_nodes["cluster"] != -1]
                            .dissolve(by="cluster")
                            .concave_hull(ratio=CONCAVE_HULL_RATIO)
                            .buffer(VERY_SMALL_BUFFER_SIZE)
                        ),
                        (
                            reached_nodes[reached_nodes["cluster"] == -1].buffer(
                                VERY_SMALL_BUFFER_SIZE
                            )
                        ),
                    ]
                ).union_all()

                isochrones["travel_time"].append(isochrone)
                isochrones["geometry"].append(isochrone_polygons)

        isochrones = geopandas.GeoDataFrame(
            isochrones, geometry="geometry", crs=self.EQUIDISTANT_CRS
        )

        # clip smaller isochrones by larger isochrones
        # (concave_hull’s ratio parameter depends on input shapes and does not
        # produce the same results, e.g., around bridges or at the coast line)
        for row in range(len(isochrones) - 2, 0, -1):
            isochrones.loc[row, "geometry"] = shapely.intersection(
                isochrones.loc[row, "geometry"], isochrones.loc[row + 1, "geometry"]
            )

        isochrones["geometry"] = (
            isochrones["geometry"]
            .buffer(CONCAVE_HULL_BUFFER_SIZE)
            .boundary.apply(
                lambda geometry: (
                    geometry
                    if isinstance(geometry, shapely.MultiLineString)
                    else shapely.MultiLineString([geometry])
                )
            )
            .apply(
                lambda multilinestring: (
                    shapely.MultiLineString(
                        [
                            simplification.cutil.simplify_coords_vwp(
                                linestring.coords,
                                self.point_grid_resolution * 5.0,
                            )
                            for linestring in multilinestring.geoms
                        ]
                    )
                )
            )
            .to_crs(R5_CRS)
        )

        return isochrones

    @property
    def destinations(self):
        """A regular grid of points covering the range of the chosen transport mode."""
        try:
            return self._destinations
        except AttributeError:
            destinations = self._regular_point_grid
            destinations["geometry"] = self.transport_network.snap_to_network(
                destinations["geometry"]
            )
            destinations = destinations[destinations["geometry"] != EMPTY_POINT]
            destinations["geometry"] = destinations["geometry"].normalize()
            destinations = destinations.drop_duplicates()

            # with snapping, sometimes we end up with clumps of points
            # below, we try to form clusters, from all clusters we retain
            # one geometry, only
            destinations = SpatiallyClusteredGeoDataFrame(
                destinations, eps=(0.5 * self.point_grid_resolution)
            )
            destinations = pandas.concat(
                [
                    (
                        destinations[destinations["cluster"] != -1]
                        .groupby("cluster")
                        .first()
                        .set_crs(R5_CRS)
                    ),
                    destinations[destinations["cluster"] == -1],
                ]
            )[["id", "geometry"]].copy()

            if self.point_grid_sample_ratio < 1.0:
                destinations = destinations.sample(frac=self.point_grid_sample_ratio)

            self._destinations = destinations

            return destinations

    @destinations.setter
    def destinations(self, destinations):
        # https://bugs.python.org/issue14965
        super(self.__class__, self.__class__).destinations.__set__(self, destinations)

    @property
    def isochrones(self):
        """
        Compute isochrones for these travel times.

        pandas.TimedeltaIndex | collections.abc.Iterable[int]
        An iterable of integers is interpreted as minutes.
        """
        try:
            return self._isochrones
        except AttributeError:
            raise

    @isochrones.setter
    def isochrones(self, isochrones):
        if not isinstance(isochrones, pandas.TimedeltaIndex):
            isochrones = pandas.to_timedelta(isochrones, unit="minutes")
        try:
            # do not compute for 0 travel time
            isochrones = isochrones.drop(datetime.timedelta(0))
        except KeyError:
            pass
        self._isochrones = isochrones

    @property
    def _regular_point_grid(self):
        extent = shapely.ops.transform(
            pyproj.Transformer.from_crs(
                R5_CRS,
                self.EQUIDISTANT_CRS,
                always_xy=True,
            ).transform,
            self.transport_network.extent,
        )

        grid = geohexgrid.make_grid_from_bounds(
            *extent.bounds,
            self.point_grid_resolution,
            crs=self.EQUIDISTANT_CRS,
        )
        grid["geometry"] = grid["geometry"].centroid
        grid["id"] = grid.index
        grid = grid[["id", "geometry"]].to_crs(R5_CRS)

        # for walking and cycling, we can clip the extent to an area reachable
        # by the (well-defined) travel speeds:
        if set(self.request.transport_modes) <= set(
            (TransportMode.WALK, TransportMode.BICYCLE)
        ):
            if TransportMode.WALK in self.request.transport_modes:
                speed = self.request.speed_walking
            if TransportMode.BICYCLE in self.request.transport_modes:
                speed = self.request.speed_cycling

            speed = speed * (1000.0 / 3600.0) * 1.1  # km/h -> m/s, plus a bit of buffer

            grid = grid.clip(
                (
                    pandas.concat([self.origins] * 2)  # workaround until
                    # https://github.com/pyproj4/pyproj/issues/1309 is fixed
                    .to_crs(self.EQUIDISTANT_CRS)
                    .buffer(speed * max(self.isochrones).total_seconds())
                    .to_crs(R5_CRS)
                )
            )

        return grid.copy()



#!/usr/bin/env python3

"""Determine a reasonable memory footprint for the Java virtual machine."""


import psutil
import re
import warnings

from .config import Config
from .warnings import R5pyWarning


__all__ = ["MAX_JVM_MEMORY"]


ABSOLUTE_MINIMUM_MEMORY = 200 * 1024**2  # never grant less than 200 MiB to JVM


config = Config()
config.argparser.add(
    "-m",
    "--max-memory",
    help="""
        Memory limit for the JVM running R5.

        Use % as a suffix to specify a share of total RAM;
        K, M, G, T to specify KiB, MiB, GiB, or TiB, respectively.
        Values without suffix are interpreted as bytes.
    """,
    default="80%",
)


def _share_of_ram(share=0.8, leave_at_least=(2 * 1024**3)):
    """
    Calculate a share of total RAM.

    Arguments
    ---------
    share : float
        Which portion of total RAM to return.
        Default: 0.8
    leave_at_least : float
        How much RAM (in bytes) to leave (for other applications and system) in any case.
        If `total RAM - (total RAM ⨉ share)` is smaller than `leave_at_least`,
        return `total RAM - leave_at_least`, instead.
        Default: 2GiB

    Returns
    -------
    int
        A value in bytes that is close to `share` portion of total RAM.
    """
    total_ram = psutil.virtual_memory().total
    if total_ram * (1.0 - share) > leave_at_least:
        share_of_ram = round(share * total_ram)
    else:
        share_of_ram = round(total_ram - leave_at_least)
    return share_of_ram


def _parse_value_and_unit(value_and_unit, max_unit_length=1):
    """
    Extract value and unit from a string.

    The string is allowed to contain a (non-numeric) unit suffix.
    For instance, input values of `'1M'` or `3.732G` would yield return
    values `(1, 'M')` or `(3.732, 'G')`, respectively.

    Arguments
    ---------
    value_and_unit : str
        Input string (typically passed-through from config
        parameter `--max-memory`).
    max_unit_length : int
        Maximum length in characters of the unit suffix.
        Default: 1

    Returns
    -------
    tuple: a tuple containing
        - value (float): The value extracted from `value_and_unit`.
        - unit (str): The unit extracted from `value_and_unit`.
    """
    matches = re.match(
        (
            "^(?P<value>[0-9]+(\\.[0-9]+)?)"  # value
            f"(?P<unit>[^0-9]){{0,{max_unit_length}}}$"  # unit
        ),
        value_and_unit,
    )
    value = float(matches["value"])
    unit = matches["unit"]

    return value, unit


def _interpret_power_of_two_units(value, unit):
    """
    Convert a value given as value and power-of-two unit into a bytes value.

    Arguments
    ---------
    value : float
        Input value.
    unit : str
        Unit suffix, as specified in IEC 80000-13, e.g., 'K' for kibibyte.

    Returns
    -------
    int:
        interpreted value in bytes
    """
    # the position of each suffix in this string is the unit’s exponent
    # over 1024.
    # Compare https://en.wikipedia.org/wiki/ISO%2FIEC_80000#Part_13:_Information_science_and_technology
    SUFFIXES = " KMGTPEZY"

    if unit is None:
        unit = " "

    if unit not in SUFFIXES:
        raise ValueError(
            f"Could not interpret unit '{unit}'. "
            "Allowed suffixes are 'K', 'M', 'G', 'T', 'P', 'E', 'Z', and 'Y'."
        )

    exponent = SUFFIXES.find(unit)
    value *= 1024**exponent

    return value


def _get_max_memory(max_memory):
    """
    Interpret the config parameter --max-memory.

    Arguments
    ---------

    max_memory : str
        Memory limit for the JVM running R5.

        Use % as a suffix to specify a share of total RAM;
        K, M, G, T suffix specify KiB, MiB, GiB, or TiB, respectively.
        Values without suffix are interpreted as bytes.

    Returns
    -------
    int
        Maximum amount of memory allocated for R5 in bytes.
    """
    try:
        value, unit = _parse_value_and_unit(max_memory)
    except TypeError:
        raise ValueError(f"Could not interpret `--max-memory` ('{max_memory}').")

    if unit == "%":
        value = _share_of_ram(share=(value / 100.0))
    else:
        # convert to bytes
        value = _interpret_power_of_two_units(value, unit)

    max_memory = round(value)

    if max_memory < ABSOLUTE_MINIMUM_MEMORY:
        max_memory = ABSOLUTE_MINIMUM_MEMORY
        warnings.warn(
            f"Requested maximum JVM heap size is too low for R5, "
            f"setting to minimum value {ABSOLUTE_MINIMUM_MEMORY:d} bytes.",
            R5pyWarning,
        )

    return max_memory


MAX_JVM_MEMORY = _get_max_memory(config.arguments.max_memory)



#!/usr/bin/env python3


"""Represent one leg of a trip."""


import datetime
import numpy
import shapely


__all__ = ["TripLeg"]


class TripLeg:
    """
    Represent one leg of a trip.

    This is a base class, use one the specific classes,
    e.g., TransitLeg, or DirectLeg
    """

    COLUMNS = [
        "transport_mode",
        "departure_time",
        "distance",
        "travel_time",
        "wait_time",
        "feed",
        "agency_id",
        "route_id",
        "start_stop_id",
        "end_stop_id",
        "geometry",
    ]

    def __init__(
        self,
        transport_mode=None,
        departure_time=numpy.datetime64("NaT"),
        distance=None,
        travel_time=datetime.timedelta(seconds=0),
        wait_time=datetime.timedelta(seconds=0),
        feed=None,
        agency_id=None,
        route_id=None,
        start_stop_id=None,
        end_stop_id=None,
        geometry=shapely.LineString(),
    ):
        """
        Represent one leg of a trip.

        This is a base class, use one of the more specific classes, e.g.,
        TransitLeg, or DirectLeg

        Arguments
        =========
        transport_mode : r5py.TransportMode
            mode of transport this trip leg was travelled
        departure_time : datetime.datetime
            departure time of this trip leg
        distance : float
            distance covered by this trip leg, in metres
        travel_time : datetime.timedelta
            time spent travelling on this trip leg
        wait_time : datetime.timedelta
            time spent waiting for a connection on this trip leg
        feed : str
            the GTFS feed identifier used for this trip leg
        agency_id : str
            the GTFS id the agency used for this trip leg
        route_id : str
            the GTFS id of the public transport route used for this trip leg
        start_stop_id : str
            the GTFS stop_id of the boarding stop used for this trip leg
        end_stop_id : str
            the GTFS stop_id of the aligning stop used for this trip leg
        geometry : shapely.LineString
            spatial representation of this trip leg
        """
        self.transport_mode = transport_mode
        self.departure_time = departure_time
        self.distance = distance
        self.travel_time = travel_time
        self.wait_time = wait_time
        self.feed = feed
        self.agency_id = agency_id
        self.route_id = route_id
        self.start_stop_id = start_stop_id
        self.end_stop_id = end_stop_id
        self.geometry = geometry

    def __add__(self, other):
        """Trip-chain `other` to `self`."""
        from .trip import Trip

        if isinstance(other, self.__class__):
            trip = Trip([self, other])
            return trip
        elif isinstance(other, Trip):
            other.legs = [self] + other.legs
            return other
        else:
            raise TypeError(
                f"unsupported operand type(s) for '+': '{type(other)}' and '{type(self)}'"
            )

    def __radd__(self, other):
        """Trip-chain `self` to `other`."""
        from .trip import Trip

        if other == 0:  # first iteration of sum()
            return self
        elif isinstance(other, Trip):
            other.legs.append(self)
            return other
        else:
            return self.__add__(other)

    def __eq__(self, other):
        """Check if `other` is an equal `TripLeg`."""
        if isinstance(other, self.__class__):
            return False not in [
                self._are_columns_equal(other, column) for column in self.COLUMNS
            ]

    def __gt__(self, other):
        """Check if `other` has a longer travel time."""
        if isinstance(other, TripLeg):
            return (self.travel_time + self.wait_time) > (
                other.travel_time + other.wait_time
            )

    def __ge__(self, other):
        """Check if `other` has a longer or equal travel time."""
        if isinstance(other, TripLeg):
            return (self.travel_time + self.wait_time) >= (
                other.travel_time + other.wait_time
            )

    def __lt__(self, other):
        """Check if `other` has a shorter travel time."""
        if isinstance(other, TripLeg):
            return (self.travel_time + self.wait_time) < (
                other.travel_time + other.wait_time
            )

    def __le__(self, other):
        """Check if `other` has a shorter or equal travel time."""
        if isinstance(other, TripLeg):
            return (self.travel_time + self.wait_time) <= (
                other.travel_time + other.wait_time
            )

    def __repr__(self):
        """Return a string representation."""
        try:
            first_point = self.geometry.coords[0]
            last_point = self.geometry.coords[-1]
            _repr = (
                "<"
                f"{self.__class__.__name__}: "
                f"{self.transport_mode}, "
                f"{self.distance}m, "
                f"{self.travel_time.total_seconds():0.1f}s, "
                f"{first_point} -> {last_point}"
                ">"
            )
        except (AttributeError, IndexError):
            _repr = f"<{self.__class__.__name__}>"
        return _repr

    def _are_columns_equal(self, other, column):
        """
        Check if a column equals the same column of a different `Trip`.

        Compare if attribute `column` of self equals attribute `column` of
        other. Also True if both values are None or NaN or NaT.
        """
        self_column = getattr(self, column)
        other_column = getattr(other, column)

        return (
            self_column == other_column
            or (self_column is None and other_column is None)
            or (self_column == numpy.nan and other_column == numpy.nan)
            or (
                self_column == numpy.datetime64("NaT")
                and other_column == numpy.datetime64("NaT")
            )
        )

    def as_table_row(self):
        """
        Return a table row (list) of this trip leg’s details.

        Returns
        =======
        list : detailed information about this trip leg: ``transport_mode``,
            ``departure_time``, ``distance``, ``travel_time``, ``wait_time``,
            ``feed``, ``agency_id`` ``route_id``, ``start_stop_id``,
            ``end_stop_id``, ``geometry``
        """
        return [getattr(self, column) for column in self.COLUMNS]



#!/usr/bin/env python3


import sys

import pytest

import r5py


class TestVerboseWarnings:
    @pytest.fixture
    def setup_verbose_mode(self):
        original_sys_argv = sys.argv.copy()
        sys.argv.append("--verbose")
        yield
        sys.argv = original_sys_argv

    def test_trip_planner_warn_if_origin_or_destination_not_found(
        self,
        transport_network,
        regional_task,
        setup_verbose_mode,
    ):
        regional_task.transport_modes = [r5py.TransportMode.WALK]

        regional_task.fromLat = 48.20
        regional_task.fromLon = 16.37
        regional_task.toLat = 0
        regional_task.toLon = -78.5

        with pytest.warns(RuntimeWarning, match="Could not find"):
            _ = r5py.r5.trip_planner.TripPlanner(
                transport_network,
                regional_task,
            ).trips

    def test_detailed_itineraries_warn_no_destinations_all_to_all(
        self,
        transport_network,
        population_grid_points_first_three,
        departure_datetime,
        setup_verbose_mode,
    ):
        with (
            pytest.warns(
                RuntimeWarning,
                match="No destinations specified, computing an all-to-all matrix",
            ),
            pytest.warns(
                RuntimeWarning,
                match="No routing destinations defined, using origins as destinations, too.",
            ),
        ):
            r5py.DetailedItineraries(
                transport_network=transport_network,
                origins=population_grid_points_first_three,
                departure=departure_datetime,
                transport_modes=[r5py.TransportMode.WALK],
            )

    def test_detailed_itineraries_warn_diff_length_all_to_all(
        self,
        transport_network,
        population_grid_points_first_three,
        population_grid_points_four,
        departure_datetime,
        setup_verbose_mode,
    ):
        with pytest.warns(
            RuntimeWarning,
            match="Origins and destinations are of different length, computing an all-to-all matrix",
        ):
            r5py.DetailedItineraries(
                transport_network=transport_network,
                origins=population_grid_points_first_three,
                destinations=population_grid_points_four,
                departure=departure_datetime,
                transport_modes=[r5py.TransportMode.WALK],
            )

    def test_travel_time_matrix_warn_no_destinations(
        self,
        transport_network,
        population_grid_points_first_three,
        departure_datetime,
        setup_verbose_mode,
    ):
        with pytest.warns(
            RuntimeWarning,
            match="No routing destinations defined, using origins as destinations, too.",
        ):
            r5py.TravelTimeMatrix(
                transport_network,
                origins=population_grid_points_first_three,
                departure=departure_datetime,
            )

    def test_detailed_itineraries_warn_origins_equal_to_destinations(
        self,
        transport_network,
        population_grid_points_first_three,
        departure_datetime,
        setup_verbose_mode,
    ):
        with pytest.warns(
            RuntimeWarning,
            match="Origins and destinations are identical, computing an all-to-all matrix",
        ):
            detailed_itineraries_computer = r5py.DetailedItineraries(
                transport_network=transport_network,
                origins=population_grid_points_first_three,
                destinations=population_grid_points_first_three,
                departure=departure_datetime,
                transport_modes=[r5py.TransportMode.WALK],
            )
        assert detailed_itineraries_computer.all_to_all
